# pylint: disable=C0325, C0103, C0111

""" Constants """
import copy
import json
import os

import pureyaml
from appdirs import user_data_dir

OPTION_PRECEDENCE = ['cli', 'env', 'ini', 'dft']

OPTION_TYPE_MAP = {
    'debug': str,
    'murano_host': str,
    'murano_id': str,
    'config_io_file': str,
    'config_applications_file': str,
    'ini_file': str,
    'murano_token': str,
    'watchlist': list,
    'http_timeout': int,
    'certfile': str,
    'pkeyfile': str,
    'murano_cacert': str,
    'murano_port': int,
    'queuing': bool,
    'queuing_batch_amount': int,
    'queuing_db_max_size': int,
    'queuing_db_path': str,
    'queuing_record_rate': int,
    'queuing_remove_unused_device': bool,
    'queuing_drop_strategy': str,
    'no-logrotate': bool,
    'log-filename': str,
    'log-max-bytes': int,
    'log-max-backups': int,
}

OPTION_NAME_MAPPER = {
    'env': lambda d, x: d.get('EDGED_{}'.format(x.upper()), None),
    'cli': lambda d, x: d.get('--{}'.format(x.replace('_', '-')), None),
    'dft': lambda d, x: d.get(x, None),
    'ini': lambda d, x: d.get(x)
}

def _ini_get(I, string):
    try:
        s, o = string.split('.')
        if s in I.sections():
            return I.get(s, o)
    except:
        return None
    return None


DEFAULTS = {
    'debug': os.environ.get('EDGED_LOG_DEBUG', 'CRITICAL').upper(),
    'config_io_file': 'config_io.json',
    'config_applications_file': 'config_applications.json',
    'watchlist': ['config_io', 'data_out', 'config_applications'],
    'http_timeout': 60*5*1000,
    'ini_file': 'edged.ini',
    'no-logrotate': False,
    'log-filename': os.environ.get('EDGED_LOG_FILENAME') or 'stdout',
    'log-max-bytes': int(os.environ.get('EDGED_LOG_MAX_BYTES', 1024*1000)),
    'log-max-backups': int(os.environ.get('EDGED_LOG_MAX_BACKUPS', 3)),
    'queuing_batch_amount': 0,
    'queuing_db_max_size': 20000000,
    'queuing_db_path': user_data_dir('exoedge', 'exosite'),
    'queuing_drop_strategy': 'oldest',
    'queuing_record_rate': 1,
    'queuing_remove_unused_device': False
}



# pylint: disable=C0103,W0212,C0111,C0325
import os
import unittest
import threading
from murano_client.client import MuranoClient
from exoedge.config_io import ConfigIO
from exoedge.config_applications import ConfigApplications

test_dir = os.path.dirname(os.path.abspath(__file__))


class TestConfigIO(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        print("setup for test: {}".format(self.id().split('.')[-1]))

    def tearDown(self):
        print("teardown for test: {}".format(self.id().split('.')[-1]))

    def test_no_local_config(self):
        c = ConfigIO(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
            )
        )
        self.assertIsNotNone(c)
        self.assertFalse(c.channels)
        self.assertEqual(c.config_file, 'config_io.json')

    def test_no_device(self):
        # config objects need MuranoClient objects.
        with self.assertRaises(AssertionError):
            ConfigIO()

    def test_default_config_file(self):
        c = ConfigIO(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
                murano_id='MYMURANOID'
            )
        )
        self.assertEqual(c.config_file, 'config_io.json')

    def test_complete_config_file(self):
        path_to_file = os.path.join(test_dir, 'assets/complete-channel-config.json')
        c = ConfigIO(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
                murano_id='MYMURANOID'
            ),
            config_io_file=path_to_file
        )
        self.assertEqual(c.config_file, path_to_file)

    def test_parse_name(self):
        c = ConfigIO(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
                murano_id='MYMURANOID'
            ),
            config_io_file={}
        )
        self.assertEqual(c.parse_name('Modbus_TCP'),
                         ('exoedge_modbus', 'ModbusExoEdgeSource'))
        self.assertEqual(c.parse_name('Modbus_RTU'),
                         ('exoedge_modbus', 'ModbusExoEdgeSource'))
        self.assertEqual(c.parse_name('CANOpen'),
                         ('exoedge_canopen', 'CanopenExoEdgeSource'))


class TestConfigApplications(unittest.TestCase):
    def test_no_local_config(self):
        c = ConfigApplications(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
            )
        )
        self.assertIsNotNone(c)
        self.assertEqual(c.config_file, 'config_applications.json')

    def test_no_device(self):
        # config objects need MuranoClient objects.
        with self.assertRaises(AssertionError):
            ConfigApplications()

    def test_default_config_file(self):
        c = ConfigApplications(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
                murano_id='MYMURANOID'
            )
        )
        self.assertEqual(c.config_file, 'config_applications.json')

    def test_complete_config_file(self):
        path_to_file = os.path.join(test_dir, 'assets/complete-channel-config.json')
        c = ConfigApplications(
            device=MuranoClient(
                murano_host='https://random.m2.exosite.io/',
                watchlist=['config_io'],
                murano_id='MYMURANOID'
            ),
            config_applications_file=path_to_file
        )
        self.assertEqual(c.config_file, path_to_file)


def main():
    unittest.main()


if __name__ == "__main__":
    main()



# pylint: disable=C0103,W0602,C0325,C0301,C0111
import os
import unittest
import json
import pureyaml
from exoedge import channel

test_dir = os.path.dirname(os.path.abspath(__file__))


class TestChannel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass
    @classmethod
    def tearDownClass(cls):
        pass
    def setUp(self):
        print("setup for test: {}".format(self.id().split('.')[-1]))
    def tearDown(self):
        print("teardown for test: {}".format(self.id().split('.')[-1]))


    def test_001_channel_constructor(self):
        c = channel.Channel(
            **json.load(open(os.path.join(test_dir, 'assets/complete-channel-config.json')))
        )
        self.assertIsNotNone(c)

    def test_002_channel_roch(self):
        ch_conf = json.load(
            open(
                os.path.join(test_dir, 'assets/complete-channel-config.json'))
        )
        print(ch_conf)
        ch_conf['protocol_config']['report_on_change'] = True
        ch_conf['protocol_config']['report_on_change_tolerance'] = 0.1
        c = channel.Channel(**ch_conf)
        self.assertIsNotNone(c.ROCH.filter_data([channel.Sample(1.0)], 1.2)[-1].data)

    def test_003_channel_roch(self):
        ch_conf = json.load(
            open(
                os.path.join(test_dir, 'assets/complete-channel-config.json'))
        )
        print(ch_conf)
        ch_conf['protocol_config']['report_on_change'] = True
        ch_conf['protocol_config']['report_on_change_tolerance'] = 0.1
        c = channel.Channel(**ch_conf)
        self.assertIsNone(c.ROCH.filter_data([channel.Sample(1.0)], 1.02)[-1].data)


def main():
    unittest.main()

if __name__ == "__main__":
    main()



# pylint: disable=C0122,C0103
# -*- coding: utf-8 -*-
import json
import pureyaml
import copy
from six import string_types

def get_nested_object(obj, chain):
    """
        Recursive algo for getting nested elements of a dictionary.

        Input: obj - a python dictionary
        Input: chain - a 'list defining the path to a dictionary element
        Returns: object specified by chain (e.g. ['fee', 'fie']) or None

        Example:
            >>> get_nested_object({'foo': {'bar':{'fee': 'fie'}}}, ['foo', 'bar'])
            {'fee': 'fie'}
            >>> get_nested_object({'foo': {'bar':{'fee': 'fie'}}}, ['foo', 'bar', 'fee'])
            fie
    """
    _key = chain.pop(0)
    if _key in obj:
        return get_nested_object(obj[_key], chain) if chain else obj[_key]

def deepupdate(defaults, update):
    """ Recursively fill in missing elements
    of dictionary <update> according to defaults in
    <defaults>.

    Function uses copy.deepcopy() to avoid inadvertently
    copying references. Everything is copied by value.
    """
    for key, value in defaults.items():
        if key not in update:
            update[key] = copy.deepcopy(value)
        elif isinstance(value, dict):
            deepupdate(copy.deepcopy(value), update[key])

####################################
#    CHANNEL NAMESPACES
####################################


class Namespace(object):
    """
        This class accepts any dictionary and creates
        a new object that can be traversed via the
        '.' (dot) operator instead of normal dictionary
        syntax (e.g. d['key'], d.get('key')).

        Example:

            >>> d = {'a':{'b':{'c':'d'}}}
            >>> ns = Namespace(**d)
            >>> ns.a.b.c
            d

        There are also methods to assist in dumping
        the new namespace to JSON and YAML formats.
    """
    def __init__(self, **kw):
        # TODO: basically, the leaf() function, takes a dict and then
        # makes a NS out of it. so, if self.__dict__ gets updated,
        # you could just call leaf() on it again. a merge() function
        # would, then, just merge the two dicts, then run through
        # the result with leaf - updating (merging) the two NS's.
        self.leaf(self, kw)

    def __iter__(self):
        for k, v in self.__dict__.items():
            if isinstance(v, Namespace):
                v = dict(v)
            yield k, v


    @classmethod
    def leaf(cls, ns, d):
        for k, v  in d.items():
            try:
                v = json.loads(json.dumps(v))
            except (TypeError, ValueError):
                pass
            if k == 'app_specific_config':
                ns.__dict__[k] = v
                continue
            if isinstance(v, dict):
                leaf_ns = Namespace(**v)
            else:
                leaf_ns = v
            setattr(ns, k, leaf_ns)


    def serialize(self):
        d = {}
        def unnamespace(ns):
            _d = {}
            for i, j in ns.__dict__.items():
                if isinstance(j, Namespace):
                    _d = unnamespace(j)
                else:
                    _d[i] = j
            return _d

        for k, v in self.__dict__.items():
            if isinstance(v, Namespace):
                d[k] = unnamespace(v)
            else:
                try:
                    d[k] = v
                    json.dumps(d)
                except TypeError:
                    d[k] = str(v)
        return d


    def to_yaml(self, indent=2):
        return pureyaml.dumps(dict(self.serialize()), sort_keys=True, allow_unicode=True, indent=indent)

    def to_json(self, indent=2):
        return json.dumps(dict(self.serialize()), indent=indent)


class ChannelNamespace(Namespace):
    """
        This class has all of the benefits of the
        Namespace class, but it fills any missing
        elements so that the edged Channel object
        can use the '.' operator in source code
        safely.
    """
    def __init__(self, **kw):
        if not kw.get('__ns_no_defaults__'):
            # assure we have a basic namespace
            deepupdate(BASIC_CHANNEL_SCHEMA_DEFAULTS, kw)
            # this key has the bare minimum for any fieldbus-type in exosense
            if kw['protocol_config'].get('application'):
                if 'Modbus_TCP' == kw['protocol_config'].get('application'):
                    deepupdate(MODBUS_TCP_CHANNEL_SCHEMA_DEFAULTS, kw)
                elif 'Modbus_RTU' == kw['protocol_config'].get('application'):
                    deepupdate(MODBUS_RTU_CHANNEL_SCHEMA_DEFAULTS, kw)
                elif 'CANOpen' == kw['protocol_config'].get('application'):
                    deepupdate(CANOPEN_CHANNEL_SCHEMA_DEFAULTS, kw)
                else:
                    deepupdate(CLASSIC_CHANNEL_SCHEMA_DEFAULTS, kw)
        elif '__ns_no_defaults__' in kw:
            kw.pop('__ns_no_defaults__')
        Namespace.__init__(self, **kw)

####################################
#    CHANNEL SCHEMAS
####################################

BASIC_CHANNEL_SCHEMA_TYPE_MAP = {
    "display_name": str,
    "description": str,
    "properties": {
        "data_type": str,
        "data_unit": str,
        "min": float,
        "max": float,
        "precision": float,
        "value_mapping": str,
        "device_diagnostic": bool,
        "data_out": bool,
    },
    "protocol_config": {
        "application": str,
        "interface": str,
        "input_raw": {
            "max": float,
            "min": float,
            "unit": str,
        },
        "multiplier": float,
        "offset": float,
        "sample_rate": float,
        "report_rate": float,
        "report_on_change": bool,
        "report_on_change_tolerance": float,
        "down_sample": str,
        "app_specific_config": {},
    }
}

BASIC_CHANNEL_SCHEMA_DEFAULTS = {
    "display_name": "not specified",
    "description": "not specified",
    "properties": {
        "data_type": "not specified",
        "data_out": False,
    },
    "protocol_config": {
        "application": "not specified",
        "interface": "not specified",
        "multiplier": 1,
        "offset": 0,
        "sample_rate": 1000,
        "report_rate": 1000,
        "report_on_change": False,
        "report_on_change_tolerance": 0.0,
        "down_sample": "ACT",
        "app_specific_config": {},
    }
}

CLASSIC_CHANNEL_SCHEMA_TYPE_MAP = {
    "protocol_config": {
        "app_specific_config": {
            "module": str,
            "function": str,
            "parameters": dict,
            "positionals": list,
        }
    }
}
CLASSIC_CHANNEL_SCHEMA_DEFAULTS = {
    "protocol_config": {
        "app_specific_config": {
            "parameters": {},
            "positionals": [],
        }
    }
}
deepupdate(
    BASIC_CHANNEL_SCHEMA_DEFAULTS,
    CLASSIC_CHANNEL_SCHEMA_DEFAULTS
)


####################################
#    MODBUS TCP CHANNEL SCHEMA
####################################

MODBUS_TCP_CHANNEL_SCHEMA_TYPE_MAP = {
    "protocol_config": {
        "application": str,
        "interface": str,
        "app_specific_config": {
            "ip_address": str,
            "port": int,
            "register_range": str,
            "register_offset": int,
            "register_count":  int,
            "byte_endianness": str,
            "register_endianness": str,
            "evaluation_mode": str,
            "bitmask": hex,
        },
    }
}

MODBUS_TCP_CHANNEL_SCHEMA_DEFAULTS = {
    "protocol_config": {
        "app_specific_config": {
            "ip_address": "127.0.0.1",
            "port": 5020,
            "register_range": "INPUT_COIL",
            "register_offset": 0,
            "register_count":  1,
            "byte_endianness": "not specified",
            "register_endianness": "not specified",
            "evaluation_mode": "not specified",
            "bitmask": 0xFFFF,
        },
    }
}
deepupdate(
    BASIC_CHANNEL_SCHEMA_DEFAULTS,
    MODBUS_TCP_CHANNEL_SCHEMA_DEFAULTS
)
####################################
#    MODBUS RTU CHANNEL SCHEMA
####################################

MODBUS_RTU_CHANNEL_SCHEMA_TYPE_MAP = {
    "protocol_config": {
        "app_specific_config": {
            "slave_id": int,
            "register_range": str,
            "register_offset": int,
            "register_count":  int,
            "byte_endianness": str,
            "register_endianness": str,
            "evaluation_mode": str,
            "bitmask": hex,
        },
    }
}

MODBUS_RTU_CHANNEL_SCHEMA_DEFAULTS = {
    "protocol_config": {
        "app_specific_config": {
            "slave_id": 1,
            "register_range": "not specified",
            "register_offset": None,
            "register_count":  None,
            "byte_endianness": "not specified",
            "register_endianness": "not specified",
            "evaluation_mode": "not specified",
            "bitmask": 0xFFFF,
        },
    }
}
deepupdate(
    BASIC_CHANNEL_SCHEMA_DEFAULTS,
    MODBUS_RTU_CHANNEL_SCHEMA_DEFAULTS
)


####################################
#    CANOPEN CHANNEL SCHEMA
####################################

CANOPEN_CHANNEL_SCHEMA_TYPE_MAP = {
    "protocol_config": {
        "app_specific_config" : {
            "node_id" : hex,
            "msg_index" : hex,
            "offset" : int,
            "data_length" : int,
            "evaluation_mode" : str,
            "bitmask" : hex
        }
    }
}

CANOPEN_CHANNEL_SCHEMA_DEFAULTS = {
    "protocol_config": {
        "app_specific_config" : {
            "node_id" : 0x01,
            "msg_index" : 0x180,
            "offset" : 0,
            "data_length" : 8,
            "evaluation_mode" : "unsigned_integer",
            "bitmask" : 0xFFFFFFFF
        }
    }
}
deepupdate(
    BASIC_CHANNEL_SCHEMA_DEFAULTS,
    CANOPEN_CHANNEL_SCHEMA_DEFAULTS
)


###########################################
#
#    CONFIG APPLICATIONS SCHEMAS
#
###########################################


###########################################
#    MODBUS RTU CONFIG APPLICATIONS SCHEMA
###########################################

MODBUS_RTU_CONFIG_APPLICATIONS_SCHEMA_DEFAULTS = {
    "application_display_name" : "Modbus RTU",
    "interfaces" : [
        {
            "interface" : "/dev/tty1",
            "baud_rate" :  115200,
            "stop_bits" : 0,
            "parity" : "even"
        }
    ],
    "channel_requirements" : {
        "slave_id" : int,
        "register_range" : [
            "INPUT_COIL",
            "HOLDING_COIL",
            "INPUT_REGISTER",
            "HOLDING_REGISTER"
        ],
        "register_offset" : int,
        "register_count" : int,
        "byte_endianness" : [
            "little",
            "big"
        ],
        "register_endianness" : [
            "little",
            "big"
        ],
        "evaluation_mode" : [
            "floating point: ieee754",
            "whole-remainder",
            "signed integer",
            "unsigned",
            "bitmask_int",
            "bitmask_bool",
            "string-ascii"
        ],
        "bitmask" : hex
    }

}

###########################################
#    MODBUS TCP CONFIG APPLICATIONS SCHEMA
###########################################

MODBUS_TCP_CONFIG_APPLICATIONS_SCHEMA_DEFAULTS = {
    "application_display_name" : "Modbus TCP",
    "interfaces" : [
        "/dev/eth0",
        "/dev/wlan0"
    ],
    "channel_requirements" : {
        "ip_address" : str,
        "port" : int,
        "register_range" : [
            "INPUT_COIL",
            "HOLDING_COIL",
            "INPUT_REGISTER",
            "HOLDING_REGISTER"
        ],
        "register_offset" : int,
        "register_count" : int,
        "byte_endianness" : [
            "little",
            "big"
        ],
        "register_endianness" : [
            "little",
            "big"
        ],
        "evaluation_mode" : [
            "floating point: ieee754",
            "whole-remainder",
            "signed integer",
            "unsigned",
            "bitmask_int",
            "bitmask_bool",
            "string-ascii"
        ],
        "bitmask" : hex,
    }
}

###########################################
#    CANOPEN CONFIG APPLICATIONS SCHEMA
###########################################

CANOPEN_CONFIG_APPLICATIONS_SCHEMA_DEFAULTS = {
    "interfaces" : [
        {
            "channel" : "canA-10",
            "bitrate" : 250000
        }
    ],
    "channel_requirements" : {
        "node_id" : hex,
        "msg_index" : hex,
        "offset": int,
        "data_length": int,
        "evaluation_mode" : [
            "REAL32",
            "INT8",
            "INT16",
            "UINT16",
            "UINT32",
            "STRING",
            "BOOLEAN"
        ],
        "bitmask" : hex
    }

}




# -*- coding: utf-8 -*-
# pylint: disable=C0103,R0902,W1202
"""
Module to generate data from arbitrary source modules as defined in config_io.
"""
import sys
import time
import json
import threading
import inspect
from exoedge import get_thread_by_name
from exoedge import logger
from exoedge.namespaces import ChannelNamespace
from exoedge.sources import ExoEdgeSource
from murano_client.client import WatchQueue, StoppableThread

LOG = logger.getLogger(__name__)


class Channel(ChannelNamespace, object):
    """
    A Channel creates Signal data for ExoSense.

    This class provides methods to get new data from a source specified in
    config_io.
    """
    def __init__(self, **kwargs):
        """
            Channel initialized and started by
            config_io.ConfigIO.add_channel()
        """
        LOG.debug("Channel kwargs: {}".format(kwargs))
        if kwargs.get('__no_derived__'):
            kwargs['__ns_no_defaults__'] = True
            kwargs.pop('__no_derived__')
        ChannelNamespace.__init__(self, **kwargs)

        # DERIVED
        if not kwargs.get('__ns_no_defaults__'):
            self.downsampler = DownSampler(self.protocol_config.down_sample) # pylint: disable=E1101
            self.ROCH = ReportOnChangeHandler(
                self.protocol_config.report_on_change,  # pylint: disable=E1101
                tolerance=self.protocol_config.report_on_change_tolerance) # pylint: disable=E1101
        elif '__ns_no_defaults__' in kwargs:
            kwargs.pop('__ns_no_defaults__')
        self.last_value = None
        self.last_report_time = 0
        self.last_sample = 0
        self.q_out = WatchQueue()
        self.q_error_out = WatchQueue()
        self.callback_timer = None
        self.source = None
        LOG.info("Channel YAML:\n{}".format(self.to_yaml()))

    def is_sample_time(self, blocking=False):
        """ Sleep for the sample rate

        # FEATURE_IMPLEMENTATION: sample_rate
        """
        sr = self.protocol_config.sample_rate / 1000.0 # pylint: disable=E1101
        if blocking:
            time.sleep(sr)
        diff = time.time() - self.last_sample
        is_sample_time = diff >= sr
        return is_sample_time

    def is_report_time(self):
        """
            Determine whether or not it is time to send queued channel data.
        """
        now = time.time()
        should_report_by = self.last_report_time + self.protocol_config.report_rate / 1000.0 # pylint: disable=E1101
        if now >= should_report_by: # pylint: disable=E1101
            return True
        return False

    def set_report_stats(self, last_value, last_report_time):
        """
            Update channel stats last_value and last_report_time.
        """
        LOG.debug(
            "stats ({}): {}, {}"
            .format(self.name, last_value, last_report_time) # pylint: disable=E1101
        )
        self.last_value = last_value
        self.last_report_time = last_report_time

    def put_sample(self, sample, timestamp=None):
        """ Place data in queue.

        In the future, this method will be switchable and optionally send
        data to gmq or SQL database, e.g. send_data_to_gmq(data)

        Parameters:
        sample:           Datapoint to be placed in queue.
        """
        LOG.info(
            "putting sample({}): {}"
            .format(self.name, sample) # pylint: disable=E1101
        )
        sample_obj = Sample(
            sample,
            gain=self.protocol_config.multiplier, # pylint: disable=E1101
            offset=self.protocol_config.offset, # pylint: disable=E1101
            timestamp=timestamp)
        self.q_out.put(sample_obj)
        self.last_sample = sample_obj.timestamp

    def put_channel_error(self, error):
        """
            Helper method for sending data on the __error channel.
        """
        LOG.info(
            "putting error: {}"
            .format(error)
        )
        self.q_error_out.put(str(error))
        self.last_sample = time.time()


class DataOutChannel(Channel):
    """
        Class for processing command and control from ExoSense
        via the data_out resource.

        The differences between a DataOutChannel and a Channel are:

         - it is never sample time
         - it is always report time
         - the `execute()` function is only implemented

        on a DataOutChannel.
    """
    resource = 'data_out'
    def __init__(self, **kwargs):
        LOG.critical("DataOutChannel kwargs: {}".format(kwargs))
        Channel.__init__(self, **kwargs)

    def execute(self, data_out_value):
        LOG.critical(
            "Channel {} executing data_out with payload: {} - source {}"
            .format(self.name, data_out_value, self.source)
        )

        if self.source:
            self.source.put_data_out(self, data_out_value)
        else:
            LOG.critical(
                "Cannot execute data_out on channel source: {}"
                .format(self.source)
            )

class Sample(tuple):
    """ Class to attach a timestamp to new data that is generated """
    def __new__(cls, data, **kwargs):
        """ Subclasses tuple. (timestamp, data) """
        timestamp = kwargs.get('timestamp') or time.time()
        offset = kwargs.get('offset')
        gain = kwargs.get('gain')
        if not isinstance(data, bool):
            try:
                data = data * gain + offset
            except TypeError:
                data = data
            # Attempt to JSON serialize the data point. If not
            # serializable, force into string.
            try:
                LOG.debug("data (json): {}".format(json.dumps(data)))
            except TypeError:
                data = str(data)
        else:
            # support bool-type data
            data = data
        new = tuple.__new__(Sample, (timestamp, data))
        setattr(new, 'timestamp' , timestamp)
        setattr(new, 'data' , data)
        LOG.debug("created new sample: {}".format(new))
        return new

    def age(self):
        """ Return age of data point in seconds """
        return time.time() - self.timestamp


class DownSampler(object):
    """
        Class to deal with down sampling Channel data

        Methods:
            max:            Maximum value in list
            min:            Minimum value in list
            sum:            Sum of values in list
            avg:            Average of values in list
            act:            Last value in list ("actual value")

        Time:
            Regardless of the downsample method
            chosen, the timestamp that will be reported
            to Murano will be the timestamp of the latest
            sample in the list provided to the
            `down_sample(data_list)` method.

            This is done because users can specify their
            own timestamps for each sample via the
            `channel.put_sample(<sample>, timestamp=<user_defined>)`
            method.

    """
    method_mapper = {
        'max': max,
        'min': min,
        'sum': sum,
        'avg': lambda ls: float(sum(ls))/len(ls),
        'act': lambda ls: ls    # No change
    }

    def __init__(self, method):
        """
        Initialized by Channel.__init__()
        """
        # TODO: remove this method check once THEMVE-2944 is fixed.
        if method == 'actual':
            method = 'ACT'
        self.method = method.lower()
        self._fn = self.method_mapper.get(self.method)
        if self._fn is None:
            LOG.critical("Couldn't map method: {}, using default: ACT".format(method))
            self._fn = self.method_mapper.get('act')

    def __repr__(self):
        return '<{}: {}>'.format(self.method, self._fn)

    def down_sample(self, channel_dq):
        """ Down sample data in the list provided.

            Assume data is list of tuples.
             - e.g. [(ts, val), (ts, val), ...]
        """
        # Create iterable copy of channel.q_out.queue contents
        data_list = list(channel_dq)

        # Clear the current queue to avoid re-processing or discarding new entries
        channel_dq.clear()
        LOG.debug(
            'Performing downsample {} on {}'
            .format(self.method, data_list)
        )

        ds_dq = []
        if self.method == 'act':
            ds_dq.extend(self._fn(data_list))
        else:
            ds_dq.append(Sample(
                self._fn([d[1] for d in data_list]),
                timestamp=data_list[-1][0]
            ))

        return ds_dq


class ReportOnChangeHandler(object):
    def __init__(self, roc, tolerance=None):
        self.roc = roc
        self.tolerance = tolerance

    def __repr__(self):
        return '<{}: {}>'.format(self.roc, self.tolerance)

    def filter_data(self, latest_samples, last):
        filtered_samples = []
        latest_ts = 0
        if self.roc and last is not None:
            # Check each sample for a changed[-enough] value
            for new_sample in latest_samples:
                if self.tolerance:
                    LOG.info(
                        "Validating whether {} <= {} <= {}..."
                        .format(
                            last - self.tolerance,
                            new_sample.data,
                            last + self.tolerance)
                    )
                    if (last - self.tolerance) <= new_sample.data <= (last + self.tolerance):
                        LOG.info("Within range; not reporting.")
                    else:
                        LOG.info("Tolerance breached; adding to report list.")
                        filtered_samples.append(new_sample)
                        last = new_sample.data
                elif last != new_sample.data:
                    LOG.info("Changed value; adding to report list.")
                    filtered_samples.append(new_sample)
                    last = new_sample.data
                latest_ts = new_sample.timestamp

        # No change(s)?
        if not filtered_samples:
            return [Sample(None, timestamp=latest_ts)]

        return filtered_samples




# pylint: disable=C0103,C0301,W1202
"""
Primary module to support edged tool.
"""
import os
import json
import time
import threading
import logging
import subprocess
from exoedge import logger
from exoedge.data_in import DataIn
from exoedge.data_queuing import Queuing
from exoedge.config_io import ConfigIO
from exoedge.configs import ConfigWatcher
from exoedge.config_applications import ConfigApplications
from exoedge.commands import Command
from exoedge.constants import OPTION_NAME_MAPPER, OPTION_TYPE_MAP, OPTION_PRECEDENCE, DEFAULTS

LOG = logger.getLogger(__name__)

class ExoCommand(Command):
    """
    Create and start ExoSense client.

    usage:
        go [options]

    options:
        -h --help           Show this screen.
    """
    Name = 'go'

    def execute(self):
        from murano_client.client import MuranoClient
        from exoedge.options_handler import OptionsHandler


        # --------------------------------------
        # OPTIONS HANDLING
        # --------------------------------------
        # The value of an option comes from, in
        # decreasing priority:
        #   1. CLI switches
        #   2. Environment variables
        #   3. INI file
        #   4. Defaults
        # --------------------------------------
        OH_ARGS = {
            'cli': self.global_args,
            'env': os.environ,
            'dft': DEFAULTS
        }

        # if not OH_ARGS['cli'].get('--queuing'):
            # OH_ARGS['cli']['--queuing'] = None
        # -----------------------------------------
        # Include INI?
        _INI_FP = OH_ARGS['cli'].get('--ini-file') or OH_ARGS['env'].get('EDGED_INI_FILE') or OH_ARGS['dft'].get('ini_file') or None
        _NO_FS = OH_ARGS['cli'].get('--no-filesystem') or OH_ARGS['env'].get('EDGED_NO_FILESYSTEM') or None
        _INI_FILE = _INI_FP and not _NO_FS
        if _INI_FILE:
            from murano_client.atomicconfigparser import atomicconfigparser as IniParser
            I = IniParser(allow_no_value=True)
            I.read(_INI_FP)

            INI_VALUES = {}
            for s in I.sections():
                LOG.info('INI section %s...', s)
                for o, v in I.items(s):
                    LOG.critical('INI option {}.{} = {}'.format(s, o, v))
                    INI_VALUES.update({o: v})
            if not ('murano_id' in INI_VALUES.keys()):
                try:
                    uuid = guess_serial_number()
                    INI_VALUES['murano_id'] = uuid
                except:
                    pass

            OH_ARGS.update({'ini': INI_VALUES})
        # -----------------------------------------


        OH = OptionsHandler(OPTION_PRECEDENCE,
                            OPTION_TYPE_MAP,
                            OPTION_NAME_MAPPER,
                            **OH_ARGS)
        PARAMETERS = OH.values
        # allow exoedge.configs module to determine these defaults.
        if '--config-io-file' not in OH_ARGS['cli'] and 'EDGED_CONFIG_IO_FILE' not in OH_ARGS['env']:
            PARAMETERS.pop('config_io_file')
        if '--config-applications-file' not in OH_ARGS['cli'] and 'EDGED_CONFIG_APPLICATIONS_FILE' not in OH_ARGS['env']:
            PARAMETERS.pop('config_applications_file')

        if PARAMETERS.get('debug'):
            LOG.critical(
                "Setting all exoedge logger levels to: {}"
                .format(PARAMETERS.get('debug'))
            )
            for lgr in logger.getExoEdgeLoggers():
                lgr.setLevel(
                    getattr(
                        logging,
                        PARAMETERS.get('debug').upper()
                        )
                    )

        LOG.info("Startup parameters: \n{}"
                 .format(json.dumps(PARAMETERS, indent=2)))

        # -----------------------------------------
        # Define MuranoClient object
        # -----------------------------------------
        device = MuranoClient(memory_queue=False, **PARAMETERS)
        device.client_activate()
        TOKEN = device.murano_token() or PARAMETERS.get('murano_token') or None
        if _INI_FILE:
            if 'device' not in I.sections():
                I.add_section('device')
            I.set('device', 'murano_token', TOKEN)
            I.write(_INI_FP)
            LOG.critical('TOKEN saved to %s', _INI_FP)
        LOG.critical('TOKEN: %s', TOKEN)

        # -----------------------------------------
        # Define resource management objects
        # -----------------------------------------
        LOG.info("Startup parameters: \n{}"
                 .format(json.dumps(PARAMETERS, indent=2)))
        config_io = ConfigIO(
            device=device,
            **PARAMETERS
        )
        LOG.critical("startup config_io_file: {}".format(config_io.config_file))
        data_queuing = None
        if PARAMETERS.get('queuing'):
            optional_temp = {
                'batch_amount': PARAMETERS.get('queuing_batch_amount'),
                'db_max_size': PARAMETERS.get('queuing_db_max_size'),
                'db_path': PARAMETERS.get('queuing_db_path'),
                'drop_strategy': PARAMETERS.get('queuing_drop_strategy'),
                'record_rate': PARAMETERS.get('queuing_record_rate'),
                'remove_unused_device': PARAMETERS.get('queuing_remove_unused_device')
            }

            # remove key when value is None
            optional = {k: v for k, v in optional_temp.items() if v is not None}
            data_queuing = Queuing(PARAMETERS["murano_host"], PARAMETERS["murano_id"], device, **optional)

        data_in = DataIn(
            config_io=config_io,
            device=device,
            data_queuing=data_queuing,
            **PARAMETERS)
        config_applications = ConfigApplications(
            device=device,
            restart_channels_event=config_io.event_new_config,
            **PARAMETERS
        )
        LOG.critical("startup config_applications_file: {}".format(config_applications.config_file))
        config_mgr_list = [
            config_io,
            config_applications
        ]

        # -----------------------------------------
        # Start Reporting
        # -----------------------------------------
        device.start_client()
        data_in.start()
        config_watcher = ConfigWatcher(
            device=device,
            config_mgr_list=config_mgr_list
        )
        config_watcher.start()

        CRITICAL_THREADS = [
            config_watcher,
            data_in,
        ]
        CRITICAL_THREADS += device.watch_threads
        CRITICAL_THREADS += config_mgr_list

        EDGED_TIMEOUT = self.global_args.get('--edged-timeout')
        def STOP_PROCESS():
            """ Callback for killing edged."""
            LOG.critical("Starting STOP_PROCESS")
            if _INI_FILE:
                ini_file = IniParser(allow_no_value=True)
                ini_file.read(_INI_FP)

                syncable_murano_client_opts = {
                    'murano_id': device.murano_id(),
                    'murano_host': device.murano_host(),
                    'murano_port': device.murano_port(),
                    'murano_token': device.murano_token(),
                    'watchlist': ','.join(DEFAULTS['watchlist'])
                }
                for sect in ini_file.sections():
                    LOG.critical('INI section {}...'.format(sect))
                    if sect == 'device':
                        for opt, val in syncable_murano_client_opts.items():
                            if not ini_file.has_option(sect, opt):
                                ini_file.set(sect, opt, val)
                                LOG.critical('Setting INI option {}.{} = {}'.format(sect, opt, val))
                            # make sure latest token is saved.
                            if opt == 'murano_token':
                                ini_file.set(sect, opt, val)

                syncable_edged_opts = {
                    'config_io_file': config_io.config_file,
                    'config_applications_file': config_applications.config_file
                }
                if not ini_file.has_section('edged'):
                    ini_file.add_section('edged')
                for opt, val in syncable_edged_opts.items():
                    if not ini_file.has_option('edged', opt):
                        ini_file.set('edged', opt, val)
                        LOG.critical('Setting INI option {}.{} = {}'.format('edged', opt, val))

                syncable_queuing_opts = {
                    'queuing_batch_amount': PARAMETERS.get('queuing_batch_amount') or DEFAULTS.get('queuing_batch_amount'),
                    'queuing_db_max_size': PARAMETERS.get('queuing_db_max_size') or DEFAULTS.get('queuing_db_max_size'),
                    'queuing_db_path': PARAMETERS.get('queuing_db_path') or DEFAULTS.get('queuing_db_path'),
                    'queuing_drop_strategy': PARAMETERS.get('queuing_drop_strategy') or DEFAULTS.get('queuing_drop_strategy'),
                    'queuing_record_rate': PARAMETERS.get('queuing_record_rate') or DEFAULTS.get('queuing_record_rate'),
                    'queuing_remove_unused_device': PARAMETERS.get('queuing_remove_unused_device') or DEFAULTS.get('queuing_remove_unused_device'),
                    'queuing': PARAMETERS.get('queuing') or 'False'
                }

                if not ini_file.has_section('queuing'):
                    ini_file.add_section('queuing')
                for opt, val in syncable_queuing_opts.items():
                    ini_file.set('queuing', opt, val)
                ini_file.write(_INI_FP)

            LOG.warning("stopping...")
            device.stop_all()
            config_watcher.stop()
            data_in.stop()
            if PARAMETERS.get('queuing'):
                data_queuing.disable_dequeue()
            running_threads = threading.enumerate()
            for thread in running_threads:
                if hasattr(thread, 'stop'):
                    LOG.critical("Stopping thread: {}".format(thread))
                    if hasattr(thread, 'is_stopped'):
                        if not thread.is_stopped():
                            thread.stop()

            for thread in running_threads:
                try:
                    thread.join(timeout=1.0)
                except RuntimeError:
                    pass
                except Exception:
                    pass

            LOG.critical('TOKEN: %s', TOKEN)
            exit(0)


        class CritWatcher(object):
            """ State holder for critical thread monitoring. """
            StartTime = None
            def critical_threads_ok(self, interval=5.0):
                """ Returns boolean of whether or not any critical thread has died. """
                if time.time() - self.StartTime >= interval:
                    self.StartTime = time.time()
                    for critical_thread in CRITICAL_THREADS:
                        LOG.info(
                            "{} is_alive: {}"
                            .format(critical_thread, critical_thread.is_alive()))
                        if not critical_thread.is_alive():
                            LOG.critical(
                                "\n{}\n\n\n\n{}\n\n\n\n{}\n"
                                .format(
                                    '%'*80,
                                    "\nCRITICAL THREAD <{}> HAS DIED!!!\n".format(critical_thread),
                                    '%'*80
                                    )
                                )
                            return False
                return True

        critwatcher = CritWatcher()
        critwatcher.StartTime = time.time()

        try:
            if EDGED_TIMEOUT:
                EDGED_TIMEOUT = float(EDGED_TIMEOUT)
                BEGINNING = time.time()
                while time.time() < BEGINNING + EDGED_TIMEOUT:
                    if not critwatcher.critical_threads_ok():
                        break
                    time.sleep(1)
                STOP_PROCESS()
            else:
                while True:
                    if not critwatcher.critical_threads_ok():
                        break
                    time.sleep(1)
                STOP_PROCESS()
        except KeyboardInterrupt:
            STOP_PROCESS()
        return

def guess_serial_number():
    try:
        return get_fwprintenv_serial_number()
    except:
        pass

def get_fwprintenv_serial_number():
    return subprocess.check_output("sudo fw_printenv serialnumber | awk -F= '{print $2}'", shell=True).rstrip().decode("utf-8")



"""
ExoEdge module. Provides high-level functionality to play nicely
with Exosite's Remote Condition Monitoring solution for Industrial IoT.
"""
# pylint: disable=W1202
import json
import time
from murano_client.client import StoppableThread, WatchQueue
from exoedge.channel import Sample
from exoedge import logger

LOG = logger.getLogger(__name__)


class DataIn(StoppableThread, object):
    """
        This class has two threads for getting outbound data
        into the data_in resource:

        * DataInWriter:
            Waits on a queue for payload objects to write to data_in.

        * DataInWriter.DataWatcher
            Monitors all channels continuously and iteratively and
            puts all channel data (including __error channel) in the
            DataInWriter queue.

        Any channel that has its 'put_sample' or 'put_channel_error'
        method called will be processed for upload here.
    """

    def __init__(self, **kwargs):
        """
        DataInWriter initialized by ExoEdge.go()

        Parameters:
            device: murano_client.client.Client() object.
            config_io: exoedge.config_io.ConfigIO() object.
        """
        StoppableThread.__init__(self, name='DataInWriter')
        self.device = kwargs.get('device')
        self.config_io = kwargs.get('config_io')
        self.queuing = kwargs.get('data_queuing', None)
        self.q_outbound_data = WatchQueue()

    def channel_data_watcher(self):
        """ Process Channel queues for data and errors. This thread is
        responsible for all channel data getting into the outbound queue.
        """
        while True:
            data_in = {}
            __error_channel = {'__error': []}
            for name, channel in self.config_io.channels.items():
                if not channel.q_out.empty():
                    # FEATURE_IMPLEMENTATION: report_rate
                    if channel.is_report_time():
                        LOG.info(
                            "Downsampling {}'s queue: {}"
                            .format(channel.name, list(channel.q_out.queue))
                        )
                        downsampled_q = channel.downsampler.down_sample(
                            channel.q_out.queue
                        )
                        # FEATURE_IMPLEMENTATION: report_on_change
                        if not channel.protocol_config.report_on_change:
                            data_in[name] = downsampled_q
                            channel.set_report_stats(downsampled_q[-1].data, downsampled_q[-1].timestamp)
                        else:
                            # FEATURE_IMPLEMENTATION: report_on_change_tolerance
                            changed_values = channel.ROCH.filter_data(downsampled_q, channel.last_value)
                            if changed_values and changed_values[-1].data:
                                LOG.debug(
                                    "report_on_change_tolerance ({}) changed: {} :: {}"
                                    .format(
                                        channel.protocol_config.report_on_change_tolerance,
                                        channel.last_value,
                                        changed_values
                                    )
                                )
                                data_in[name] = changed_values
                                channel.set_report_stats(changed_values[-1].data, changed_values[-1].timestamp)
                            else:
                                LOG.warning(
                                    "report_on_change_tolerance ({}) _unchanged_ {} :: {} sec"
                                    .format(
                                        channel.protocol_config.report_on_change_tolerance,
                                        changed_values,
                                        downsampled_q[-1].timestamp - channel.last_report_time
                                    )
                                )

                while not channel.q_error_out.empty():
                    error = channel.q_error_out.safe_get(timeout=0.001)
                    if error:
                        __error_channel['__error'].append({name: error})

            if data_in:
                self.q_outbound_data.put(data_in)
            if __error_channel.get('__error'):
                self.q_outbound_data.put(__error_channel)
            time.sleep(0.01)


    def run(self):
        """
            Process the outbound queue that DataInWriter.DataWatcher
            fills with channel data and errors.
        """
        LOG.debug('starting')
        data_watcher = StoppableThread(
            name="DataInWriter.DataWatcher",
            target=self.channel_data_watcher
        )
        data_watcher.setDaemon(True)
        data_watcher.start()

        # keep the logs less noisy
        last_debug_msg = 0.0
        while not self.is_stopped():

            data = self.q_outbound_data.safe_get(timeout=0.5)
            LOG.info("Got samples from queue: {}".format(data))
            if data:
                data_in = build_data_ins(data)
                for timestamp, data_group in data_in.items():
                    if (timestamp == "no_timestamp"):
                        timestamp = time.time()
                        
                    if self.queuing:
                        self.queuing.tell(
                            resource='data_in',
                            timestamp=timestamp,
                            payload=json.dumps(data_group, ensure_ascii=False)
                        )
                    else:
                        self.device.tell(
                            resource='data_in',
                            timestamp=timestamp,
                            payload=json.dumps(data_group, ensure_ascii=False)
                        )
            else:
                now = time.time()
                if now - last_debug_msg >= 5.0:
                    LOG.debug(
                        'No data to send.'
                    )
                    last_debug_msg = now

        LOG.debug('exiting')

"""
    Takes in data samples
    Combines them based on timestamps
    Returns grouped data points

    Samples are grouped by rounding to the nearest second
"""
def build_data_ins(data):
    data_in = {}

    for channel_id, samples in data.items():
        try:
            LOG.critical( 'Parsing data ({}): {}' .format(channel_id, samples))

            for sample in samples:
                # Channel data (Sample)
                if isinstance(sample, Sample):
                    key = round(sample.timestamp) # Note: this limits channel data to second resolution (and so can result in data loss)
                    val = sample.data
                # Channel __error data (list)
                # TODO: Ensure batched __error data is updated to conform with the schema:
                # https://docs.exosite.io/schema/channel-signal_io_schema/#channel-error-handling
                else:
                    key = "no_timestamp"
                    val = sample

                if key not in data_in:
                    data_in[key] = {}

                data_in[key][channel_id] = val
        except Exception as exc:
            LOG.critical('Failed to build data_in payload: {}'.format(exc))

    return data_in



"""
pip install pureyaml
python -c "
import unittest, pureyaml
print(pureyaml.dump(sorted([e for e in dir(unittest.TestCase) if not e.startswith('_')])))
"
- addCleanup
- addTypeEqualityFunc
- assertAlmostEqual
- assertAlmostEquals
- assertDictContainsSubset
- assertDictEqual
- assertEqual
- assertEquals
- assertFalse
- assertGreater
- assertGreaterEqual
- assertIn
- assertIs
- assertIsInstance
- assertIsNone
- assertIsNot
- assertIsNotNone
- assertItemsEqual
- assertLess
- assertLessEqual
- assertListEqual
- assertMultiLineEqual
- assertNotAlmostEqual
- assertNotAlmostEquals
- assertNotEqual
- assertNotEquals
- assertNotIn
- assertNotIsInstance
- assertNotRegexpMatches
- assertRaises
- assertRaisesRegexp
- assertRegexpMatches
- assertSequenceEqual
- assertSetEqual
- assertTrue
- assertTupleEqual
- assert_
- countTestCases
- debug
- defaultTestResult
- doCleanups
- fail
- failIf
- failIfAlmostEqual
- failIfEqual
- failUnless
- failUnlessAlmostEqual
- failUnlessEqual
- failUnlessRaises
- failureException
- id
- longMessage
- maxDiff
- run
- setUp
- setUpClass
- shortDescription
- skipTest
- tearDown
- tearDownClass

"""


# pylint: disable=W0141,W0110,C0103,W1202
"""
    ExoEdge ConfigIO module to handle Channel creation from a config_io object.
"""
import inspect
import json
import os
import time

import six
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from ruamel.yaml import YAML

from exoedge import logger
from exoedge.channel import Channel, DataOutChannel
from exoedge.configs import ExoEdgeConfig
from exoedge.constants import DEFAULTS
from exoedge.namespaces import get_nested_object
from exoedge.sources import ExoEdgeClassicSource, ExoEdgeSource

LOG = logger.getLogger(__name__)


class ConfigIO(ExoEdgeConfig, object):
    """
    This class provides methods to create channels from a
    config_io and to format the data_in object to be sent
    to Murano.

    Required Inputs:
        device: a murano_client.client.Client() object
                for the connection to Murano.

    Optional Inputs:
        config:         a config_io object (dictionary).
        config_io_file: path to a JSON file containing
                        the config_io object.
        wait_timeout:   timeout, in seconds, between
                        printing 'no new config' to
                        the log (default:1.0).

    """
    resource = 'config_io'
    def __init__(self, **kwargs):
        ExoEdgeConfig.__init__(
            self,
            name="ConfigIO",
            device=kwargs.get('device'),
            config_file=kwargs.get(
                'config_io_file',
                DEFAULTS['config_io_file']
            )
        )
        self.wait_timeout = float(kwargs.get('wait_timeout', 1.0))
        self.set_config(kwargs.get('config'))
        # schema: {"$channel_id": <Channel instance>}
        self.channels = {}
        # schema: {"$source_name": {"source": <Source instance>, "channels": [ch1, chN]}}
        self.sources = {}

    def stop(self):
        """ Stop all activity.

        Stops activity of all Channel threads, Channel-
        Watcher threads, and the ConfigIO thread.
        """
        LOG.warning("Stopping ConfigIO...")
        self.stop_all_sources()
        super(ConfigIO, self).stop()

    def add_channel(self, name, ch_cfg):
        """ Create Channel and DataOutChannel objects.

        Create a Channels and DataOutChannels from a channel
        configuration object. Store them in the ConfigIO.channels
        dictionary.

        Parameters:
        name:       the key of the ch_cfg object in config_io
        ch_cfg:     the object representation of the channel
                    taken from config_io.

        """
        LOG.info(
            'Adding channel: {}'
            .format(name)
        )
        ch_cfg['name'] = name
        if get_nested_object(ch_cfg, ['properties', 'control']):
            self.channels[name] = DataOutChannel(**ch_cfg)
        else:
            self.channels[name] = Channel(**ch_cfg)
        return self.channels[name]

    @classmethod
    def is_proper_source(cls, source):
        """
            Checks source objects to ensure they're
            a subclass of ExoEdgeSource.
        """
        return isinstance(source, ExoEdgeSource)

    def add_source(self, channel):
        """
            ExoEdgeSource Rules:
             - protocol_config.application
               name/import rules
             - programming rules

            0.  Sources must subclass exoedge.sources.ExoEdgeSource.
            1.  Lowercased, the module name must be
                prefixed by "exoedge_".
            2.  Titlecased, the class name must be
                suffixed by "ExoEdgeSource".
            3.  Modbus_TCP and _RTU are the exception,
                they get stripped to just Modbus.
            4. Source class name must not be "ExoEdgeSource".

            i.e.
                CANOpen -> exoedge_canopen.CanopenExoEdgeSource
                Modbus_TCP -> exoedge_modbus.ModbusExoEdgeSource
                Phidgets -> exoedge_phidgets.PhidgetsExoEdgeSource
        """
        application = channel.protocol_config.application

        # Application default is None. When application is string, means user specify it.
        if isinstance(application, six.string_types):
            module_name, class_name = self.parse_name(application)

            # e.g. exoedge_modbus.ModbusExoEdgeSource,
            #      exoedge_canopen.CanopenExoEdgeSource,
            #      exoedge_simulator.SimulatorExoEdgeSource
            #      etc.
            source_name = '.'.join([module_name, class_name])

            # if source already loaded, append channel and return
            if self.is_source_loaded(source_name):
                loaded_source = self.sources.get(source_name)
                LOG.warning(
                    "Source '{}' already loaded."
                    .format(loaded_source)
                )
                self.sources[source_name]["channels"].append(channel)
                # keep reference to source in channel object
                channel.source = self.sources[source_name]["source"]
                return True

            if self.load_module_by_module_name(module_name, class_name, source_name, channel, application):
                loaded_source = self.sources.get(source_name)
                LOG.critical(
                    "Created source instance:: {}"
                    .format(loaded_source)
                )
                return True
            else:
                # if we've made it here we tried to find an installed ExoEdgeSource
                # but didn't find any so now try to see if we can run the channel
                # in classic style
                LOG.critical(
                    "no supported source found. trying classic module.function(*args, **kwargs) style.")
                return self.load_module_by_classic_style(channel)
        else:
            return self.load_module_by_classic_style(channel)

    def is_source_loaded(self, source_name):
        loaded_source = self.sources.get(source_name)

        return loaded_source and self.is_proper_source(loaded_source["source"])

    def parse_name(self, application):
        # Modbus_ is a special case because exosense already define it.
        if application.startswith(u'Modbus_'):
            module_name = 'exoedge_modbus'
            class_name = 'ModbusExoEdgeSource'
        else:
            # e.g. exoedge_canopen, etc.
            module_name = "exoedge_" + application.lower()
            # e.g. CanopenExoEdgeSource, etc.
            class_name = application.lower().capitalize() + "ExoEdgeSource"

        return module_name, class_name

    def load_module_by_module_name(self, module_name, class_name, source_name, channel, application):
        LOG.critical(
            "Loading module {}..."
            .format(module_name)
        )
        try:
            module = __import__(module_name)
        except Exception as exc:
            LOG.error(
                "Unable to import module {}: {}"
                .format(module_name, exc)
            )
            return False

        LOG.info("Imported module: {}".format(module))
        if hasattr(module, class_name):
            Klass = getattr(module, class_name)
            if not inspect.isclass(Klass):
                return False

            # check Klass against rules in docstring
            is_proper_class = issubclass(Klass, ExoEdgeSource) \
                                and \
                                class_name != 'ExoEdgeSource'

            if not is_proper_class:
                return False

            if self.sources.get(source_name):
                return True

            # create instance or just add a channel to existing one
            LOG.critical(
                "Creating instance: {}..."
                .format(source_name + '()')
            )
            # create source
            self.sources[source_name] = {
                "source": Klass(),
                "channels": [channel]
            }
            # keep reference to source in channel object
            channel.source = self.sources[source_name]["source"]
            return True
        else:
            LOG.error(
                "Unable to find source: {}. Putting channel {} in error."
                .format(source_name, channel)
            )
            channel.put_channel_error(
                "Unable to load source: {}"
                .format(source_name)
            )

            return False

    def load_module_by_classic_style(self, channel):
        module_name = channel.protocol_config.app_specific_config.get('module')
        function_name = channel.protocol_config.app_specific_config.get('function')
        source_name = "ExoEdgeClassicSource->" + channel.name
        try:
            module = __import__(module_name)
        except Exception as exc:
            module = None
            error_msg = "Unable to import module {}: {}".format(module_name, exc)
            LOG.error(error_msg)
            channel.put_channel_error(error_msg)
            return False

        if not hasattr(module, function_name):
            error_msg = "Module {} has no function: {}".format(module_name, function_name)
            LOG.error(error_msg)
            channel.put_channel_error(error_msg)
            return False

        if not self.sources.get(source_name):
            self.sources[source_name] = {"source": None, "channels": []}

        self.sources[source_name]["source"] = ExoEdgeClassicSource(
            channel,
            getattr(module, function_name)
        )
        self.sources[source_name]["channels"].append(channel)
        # keep reference to source in channel object
        channel.source = self.sources[source_name]["source"]

        return True

    def stop_all_sources(self):
        """ Stop ExoEdge sources.

            Call source.stop_source() for all sources.
            Called by self.run() when a new config is received.
        """
        LOG.warning('Stopping all sources!')
        for source_name in self.sources:
            source = self.sources[source_name].get("source")
            if source:
                LOG.warning(
                    "Stopping source: {}"
                    .format(source)
                )
                source.stop_source()
                if isinstance(source, ExoEdgeSource):
                    source.join(0.5)
                elif isinstance(source, ExoEdgeClassicSource):
                    # wait for one interval for graceful shutdown
                    time.sleep(source.interval)
                else:
                    LOG.critical(
                        "Don't know how to stop source: <({} ){}>"
                        .format(type(source), source)
                    )

        self.sources = {}
        self.channels = {}

    def run(self):
        """
        Update ConfigIO and Channels upon new config(s).

         - Waits for a new config_io event set by ConfigIO.set_config().
         - Stops existing sources, deletes channels before creating new ones
           based on received config object(s).
         - Starts Channels and Sources.
        """
        LOG.debug('starting')

        schema = read_yaml_schema('config_io_schema.yaml')
        while not self.is_stopped():
            self.event_new_config.clear()
            if not self.event_new_config.wait(self.wait_timeout):
                LOG.debug('no new config')
                continue

            with self._lock:
                if not self.config:
                    LOG.info("No new config_io received.")
                    continue

                LOG.debug('Received config_io')

                try:
                    validate(instance=self.config, schema=schema)
                except ValidationError as err:
                    LOG.error(err)
                    continue

                if self.sources:
                    self.stop_all_sources()

                if 'channels' in self.config.keys():
                    for name, cfg in self.config['channels'].items():
                        if 'protocol_config' in cfg:
                            self.add_channel(name, cfg)
                    for channel in self.channels.values():
                        self.add_source(channel)

                # start all sources
                LOG.info("Loaded sources: {}".format(self.sources))
                for source in self.sources:
                    LOG.warning("STARTING SOURCE: {}".format(source))
                    self.sources[source]["source"].start_source()

                LOG.critical("Running sources: {}".format(self.sources))

        LOG.debug('exiting')


def read_yaml_schema(filename):
    """
        return dict type yaml schema
    """
    file_path = os.path.realpath(
        os.path.join(os.path.dirname(__file__), filename))

    with open(file_path, 'r') as file:
        yaml = YAML(typ='safe')
        return yaml.load(file)



# pylint: disable=W1202,C0103
import sys
import threading
import time
from murano_client.client import StoppableThread, WatchQueue
from exoedge import logger

LOG = logger.getLogger(__name__)


class ExoEdgeClassicSource(object):
    def __init__(self, channel, module_function):
        self.channel = channel
        self.function = module_function
        self.interval = channel.protocol_config.sample_rate / 1000.0 # pylint: disable=E1101,
        self.timer = None

    def start_source(self):
        positionals = self.channel.protocol_config.app_specific_config.get('positionals')
        parameters = self.channel.protocol_config.app_specific_config.get('parameters')
        LOG.warning(
            "call_again: calling: {} with: {} and {}"
            .format(self.function, positionals, parameters)
        )
        try:
            self.timer = threading.Timer(
                self.interval,
                self.start_source
            )
            if positionals and parameters:
                res = self.function(
                    *positionals,
                    **parameters
                )
            elif positionals and not parameters:
                res = self.function(
                    *positionals
                )
            elif not positionals and parameters:
                res = self.function(
                    **parameters
                )
            else:
                res = self.function()

            self.channel.put_sample(res)

            self.timer.start()
        except Exception as exc:
            error_msg = "Stopping source! Calling {} raises: {}".format(
                self.function, exc)
            LOG.error(error_msg)
            self.channel.put_channel_error(error_msg)
            self.stop_source()

    def stop_source(self):
        LOG.critical("Stopping!")
        if self.timer:
            self.timer.cancel()


class DataOutQueueObject(object):
    def __init__(self, channel, data_out_value):
        self.channel = channel
        self.data_out_value = data_out_value

    def __repr__(self):
        return "<{}: {}>".format(self.channel.name, self.data_out_value)

class ExoEdgeSource(StoppableThread):
    """
        Template class for creating custom ExoEdge sources.

        To use this class, simply import and subclass it to gain access
        to the members and methods available.
    """
    _Q_DATA_OUT = WatchQueue()

    def __init__(self, **kwargs):
        """

        """
        t_kwargs = {}
        t_kwargs.update(name=kwargs.get('name') or self.__class__.__name__)
        if kwargs.get('group'):
            t_kwargs.update(group=kwargs.get('group'))
        if kwargs.get('target'):
            t_kwargs.update(target=kwargs.get('target'))
        if kwargs.get('args'):
            t_kwargs.update(args=kwargs.get('args'))
        if kwargs.get('kwargs'):
            t_kwargs.update(kwargs=kwargs.get('kwargs'))

        LOG.info("{} source thread kwargs: {}".format(self.__class__.__name__, t_kwargs))
        LOG.info("{} source other kwargs: {}".format(self.__class__.__name__, kwargs))

        super(ExoEdgeSource, self).__init__(**t_kwargs)
        self.setDaemon(True)

        # exoedge sources can have access to config_io
        self.config_io_thread = None

        # exoedge sources need access to config_applications
        self.config_applications_thread = None
        LOG.info("{} init complete.".format(str(self)))
        self.start_source = self.get_source

    def get_channels_by_application(self, application):
        """
            Function for iterating over all channels and
            compiiing a list of channels that are configured
            for
                config_io
                  .$(id)
                    .protocol_config
                      .application

            equal to the 'application' parameter.
        """
        channels_by_application = []
        if self.config_io_thread:
            for c in self.config_io_thread.channels.values():
                if c.protocol_config.application == application:
                    if c not in channels_by_application:
                        channels_by_application.append(c)
        return channels_by_application

    def get_channels_by_module(self, module):
        """
            Function for iterating over all channels and
            compiiing a list of channels that are configured
            for
                config_io
                  .$(id)
                    .protocol_config
                      .app_specific_config
                        .module

            equal to the 'module' parameter.
        """
        channels_by_module = []
        if self.config_io_thread:
            for c in self.config_io_thread.channels.values():
                if c.protocol_config.app_specific_config.get('module') == module:
                    if c not in channels_by_module:
                        channels_by_module.append(c)
        return channels_by_module

    def get_configured_applications(self):
        """
            If a value in config_applications is set or there is a local
            config_applications file, return the dictionary object.
        """
        if self.config_applications_thread and self.config_applications_thread.config:
            return self.config_applications_thread.config
        return {}

    def put_data_out(self, channel, data_out_value):
        self._Q_DATA_OUT.put(DataOutQueueObject(channel, data_out_value))

    def stop_source(self):
        LOG.warning("Stopping source: {}".format(self))
        self.stop()

    def get_source(self):
        """ Call this function atleast once to start the
        ExoEdgeSource thread. Call this function and assign
        to a variable to gain access to members and methods
        associated with this thread. """

        if not self.is_started():
            LOG.info("{} starting...".format(self.__class__.__name__))
            while not self.config_io_thread:
                for thread in threading.enumerate():
                    if thread.getName() == "ConfigIO":
                        self.config_io_thread = thread
                        break
                else:
                    LOG.critical("Waiting for config_io thread to start...")
                    time.sleep(0.25) # only executed if the inner loop did NOT break
            LOG.info("{} started.".format("ConfigIO"))
            while not self.config_applications_thread:
                for thread in threading.enumerate():
                    if thread.getName() == "ConfigApplications":
                        self.config_applications_thread = thread
                        break
                else:
                    LOG.critical("Waiting for config_applications thread to start...")
                    time.sleep(0.25) # only executed if the inner loop did NOT break
            LOG.info("{} started.".format("ConfigApplications"))
            super(ExoEdgeSource, self).start()

        LOG.warning("source: {} startup complete.".format(self.name))
        return self



import pytest

from exoedge.configs import ExoEdgeConfig
from murano_client.client import MuranoClient


@pytest.fixture(params=['https', 'mqtt'])
def muranoClient(request):
    return MuranoClient(murano_host=request.param+"://dne.m2.exosite.io/",
                          watchlist=['data_out'])

@pytest.fixture(params=[True, False], ids=["ack_successful", "ack_failed"])
def exoEdgeConfig(mocker, request, muranoClient):
    # mocker comes from pytest-mock
    # request comes from params

    mocker.patch.object(muranoClient, 'ack', return_value=(
        request.param, 'result: code: 204, body: \'No content\', success: True, authorized: True'))
    ExoEdgeConfig.resource = 'config_io'
    config = ExoEdgeConfig(
        name="ConfigIO",
        device=muranoClient
    )
    return config


def test_ack_config(exoEdgeConfig):
    assert exoEdgeConfig.acktime == 0.0
    assert exoEdgeConfig.ack_config() == None
    assert not exoEdgeConfig.acktime == 0.0



# pylint: disable=C0103,W0602,C0325,C0301,C0111
import os
import unittest
import json
import pureyaml
from exoedge import namespaces

test_dir = os.path.dirname(os.path.abspath(__file__))


class TestChannelNamespace(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass
    @classmethod
    def tearDownClass(cls):
        pass
    def setUp(self):
        print("setup for test: {}".format(self.id().split('.')[-1]))
    def tearDown(self):
        print("teardown for test: {}".format(self.id().split('.')[-1]))

    def test_001_constructor(self):
        d = {}
        n = namespaces.ChannelNamespace(**d)
        self.assertIsInstance(n, namespaces.ChannelNamespace)

    def test_002_simple_namespace(self):
        d = {'a': 'b'}
        n = namespaces.ChannelNamespace(**d)
        self.assertTrue(hasattr(n, 'a'))

    def test_003_ns_no_defaults(self):
        d = {'a': 'b', '__ns_no_defaults__': True}
        n = namespaces.ChannelNamespace(**d)
        self.assertTrue(hasattr(n, 'a'))
        self.assertFalse(hasattr(n, 'protocol_config'))
        self.assertFalse(hasattr(n, 'report_rate'))

    def test_004_print_json(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/complete-channel-config.json')).read())
        n = namespaces.ChannelNamespace(**chconf)
        tojson = n.to_json(indent=2)
        print(tojson)
        self.assertIsInstance(json.loads(tojson), dict)

    def test_005_print_yaml(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/complete-channel-config.json')).read())
        print(chconf)
        n = namespaces.ChannelNamespace(**chconf)
        toyaml = n.to_yaml()
        print(toyaml)
        print("type: {}".format(type(toyaml)))
        self.assertIsInstance(pureyaml.loads(toyaml), dict)

    def test_006_walk_namespace(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/complete-channel-config.json')).read())
        n = namespaces.ChannelNamespace(**chconf)
        self.assertIsInstance(dict(n), dict)
        self.assertTrue(hasattr(n.protocol_config, 'report_rate'))

    def test_007_config_with_missing_display_name_elem_no_defaults(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/missing-display-name.json')).read())
        chconf['__ns_no_defaults__'] = True
        n = namespaces.ChannelNamespace(**chconf)
        self.assertTrue(not hasattr(n, 'display_name'))

    def test_008_config_with_missing_display_name_elem_add_defaults(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/missing-display-name.json')).read())
        n = namespaces.ChannelNamespace(**chconf)
        self.assertTrue(hasattr(n, 'display_name'))

    def test_009_config_with_missing_properties_data_type_elem_no_defaults(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/missing-properties-data_type.json')).read())
        chconf['__ns_no_defaults__'] = True
        n = namespaces.ChannelNamespace(**chconf)
        self.assertTrue(hasattr(n, 'properties'))
        self.assertTrue(not hasattr(n.properties, 'data_type'))

    def test_010_config_with_missing_properties_data_type_elem_add_defaults(self):
        chconf = json.loads(open(os.path.join(test_dir, 'assets/missing-properties-data_type.json')).read())
        # chconf['__ns_no_defaults__'] = True
        n = namespaces.ChannelNamespace(**chconf)
        self.assertTrue(hasattr(n, 'properties'))
        self.assertTrue(hasattr(n.properties, 'data_type'))


def main():
    unittest.main()

if __name__ == "__main__":
    main()



import json
import time

import pytest
import requests

from exoedge.data_queuing import DB, Queuing
from murano_client.client import MuranoClient


@pytest.fixture
def device():
    device = MuranoClient(**{'murano_host': 'https://xyu7kqqdgf0w0000.m2.exosite.io',
                             'murano_id': 'demo',
                             'murano_token': '012b8f2357699d37a723ce0ff2a1d6d75e615f6f',
                             'watchlist': ['data_in'],
                             'murano_port': 443,
                             'memory_queue': False
                             }
                          )
    device.start_client()
    return device


@pytest.fixture
def queuing(device, mocker):
    device.outbound_protocol = 'https'
    queuing = Queuing('https://xyu7kqqdgf0w0000.m2.exosite.io/',
                      'j1939', device)
    return queuing


def test_Retention_by_real_device(queuing, device):
    data = json.dumps({'test': time.time()})
    queuing.tell('data_in', time.time(), data)
    res = device.http_read(['data_in'])
    assert res.body == 'data_in={}'.format(data)


def test_consume_table(mocker, device):
    device.outbound_protocol = 'https'
    mocker.patch.object(DB, 'read_table', return_value=[])
    queuing = Queuing(
        'https://xyu7kqqdgf0w0000.m2.exosite.io/', 'j1939', device)
    # can break loop when table is empty
    queuing.enable_dequeue()
    queuing.db.read_table.assert_called_once_with(amount=0, auto=True)
    assert queuing.dequeue_enabled == False

    # Should call https when table has data
    mocker.patch.object(queuing, 'https')
    mocker.patch.object(DB, 'read_table', return_value=[{}])
    queuing.enable_dequeue()
    queuing.https.assert_called_once_with([{}])


def test_product_id(queuing):
    assert queuing.product_id(
        'mqtt://xyu7kqqdgf0w0000.m2.exosite.io/') == 'xyu7kqqdgf0w0000'
    assert queuing.product_id(
        'https://xyu7kqqdgf0w0000.m2.exosite.io/') == 'xyu7kqqdgf0w0000'
    assert queuing.product_id('https://abc.m2.exosite.io') == 'abc'
    with pytest.raises(AttributeError):
        queuing.product_id('abc')


@pytest.mark.parametrize('protocol', ['https', 'mqtt'])
def test_tell(mocker, protocol, device):
    timestamp = time.time()
    data = json.dumps({'test': timestamp})
    device.outbound_protocol = protocol
    queuing = Queuing(protocol+'://xyu7kqqdgf0w0000.m2.exosite.io/',
                      'j1939', device)
    # test table is empty
    queuing.is_table_empty = True
    mocker.patch.object(device, 'tell', return_value=(True, ''))
    queuing.tell(resource='data_in', timestamp=timestamp, payload=data)
    device.tell.assert_called_once_with(payload=data, resource='data_in', timestamp=timestamp)

    # test table is empty and https/mqtt sent data fail
    queuing.is_table_empty = True
    mocker.patch.object(device, 'tell', return_value=(False, ''))
    mocker.patch.object(queuing.db, 'insert_record')
    queuing.tell('data_in', timestamp, data)
    queuing.db.insert_record.assert_called_once_with(timestamp, data)

    # test table is not empty
    queuing.is_table_empty = False
    mocker.patch.object(queuing.db, 'insert_record')
    queuing.tell('data_in', timestamp, data)
    queuing.db.insert_record.assert_called_once_with(timestamp, data)


class TestBD():
    @pytest.fixture
    def db(self, mocker):
        return DB('/tmp', 'queue.producid.deviceid', '5000000', 'oldest')

    def test_limit_data_size(self, mocker, db):
        # auto=True
        mocker.patch.object(db.table, 'read', return_value=[
                            (1, 1559628846.9537978, '1234567890')])
        res = db.limit_data_size(auto=True)
        # Single record is 10 byte (1234567890), '1559628846.9537978%3D1234567890%26' length is 34
        # 64000/34 = 1882.3529411764705
        assert len(res) == 1882

        # auto=False, amount = 100
        mocker.patch.object(db.table, 'read', return_value=[
                            (1, 1559628846.9537978, '1234567890') for i in range(100)])
        res = db.limit_data_size(auto=False, amount=100)
        assert len(res) == 100

        # auto=False, amount = 9999
        mocker.patch.object(db.table, 'read', return_value=[
                            (1, 1559628846.9537978, '1234567890') for i in range(9999)])
        res = db.limit_data_size(auto=False, amount=9999)
        assert len(res) == 1882


    def test_the_oldest_record_table(self, mocker, db):
        mocker.patch.object(db.table,
                            'all_table_name',
                            return_value=['queue.abcdcdkk2ls00000.device',
                                          'queue.abcdcdkk2ls00000.device2',
                                          'queue.abcdcdkk2ls00000.device3',
                                          'queue.abcdcdkk2ls00000.device4'])
        mocker.patch.object(db.table, 'query_db',
                            side_effect=[9999999, 666, 2, 1])
        res = db.table.the_oldest_record_table()
        assert res == ('queue.abcdcdkk2ls00000.device4', 1)

    def test_drop_oldest_10mins_record_and_data_in_last_10_mins(self, mocker, db):
        mocker.patch.object(db.table, 'the_oldest_record_table',
                            return_value=('queue.abcdcdkk2ls00000.device4', time.time()))
        res = db.table.drop_oldest_records()
        assert res == False

    def test_delete_unused_table(self, mocker, db):
        mocker.patch.object(db.table, 'all_table_name',
                            return_value=['queue.abcdcdkk2ls00000.device',
                                          'queue.abcdcdkk2ls00000.device2',
                                          'queue.abcdcdkk2ls00000.device3',
                                          'queue.abcdcdkk2ls00000.device4'])
        mocker.patch.object(db.table, 'delete_table', return_value=[])
        res = db.delete_unused_table()
        assert res == True



# pylint: disable=W1202
"""
Namespace for ExoEdge commands.

"""

import sys
import logging
import importlib
import pkg_resources
import docopt
from exoedge import logger

LOG = logger.getLogger(__name__)


class Command(object):
    """
        Base class for ExoEdge commands.
    """
    Name = 'go'
    def __init__(self, command_args, global_args):
        """
        Initialize the commands.
        :param command_args: arguments of the command
        :param global_args: arguments of the program
        """
        self.args = docopt.docopt(self.__doc__, argv=command_args, options_first=True)
        global_args.update({k:v for k, v in self.args.items() if v})
        self.global_args = global_args

        if isinstance(self.global_args.get('--debug'), str):
            LOG.critical("Setting Level: {}".format(self.global_args.get('--debug')))
            LOG.setLevel(getattr(logging, self.global_args.get('--debug').upper()))




# pylint: disable=W0141,W0110,C0103,W1202
"""
ExoEdge ConfigApplications.
"""
import time
import threading
import json
from exoedge.configs import ExoEdgeConfig
from exoedge import logger
from exoedge.constants import DEFAULTS

LOG = logger.getLogger(__name__)


class ConfigApplications(ExoEdgeConfig, object):
    """
        Thread for maintaining the state of the of the
        config_applications resource according to the
        exoedge.configs.ExoEdgeConfig state machine.

        If an event_new_config is detected, this will
        force the ConfigIO thread to restart all channels.
    """
    resource = 'config_applications'
    def __init__(self, **kwargs):
        ExoEdgeConfig.__init__(
            self,
            name="ConfigApplications",
            device=kwargs.get('device'),
            config_file=kwargs.get(
                'config_applications_file',
                DEFAULTS['config_applications_file']
            )
        )
        self.restart_channels_event = kwargs.get('restart_channels_event')
        self.wait_timeout = float(kwargs.get('wait_timeout', 1.0))
        self.set_config(kwargs.get('config'))

    def run(self):
        """ Update ConfigApplications.

            Waits for a new config_applications event set by
            ConfigApplications.set_config_applications().

            When a new config_applications is recieved,
            the config_io 'event_new_config' event is set
            which restarts all channels.
        """
        LOG.debug('starting')

        while not self.is_stopped():
            if self.event_new_config.wait(timeout=0.5):
                self.restart_channels_event.set()
                self.event_new_config.clear()

        LOG.debug('exiting')





# pylint: disable=W0141,W0110,C0103,W1202
import os
import threading
import json
import time
from murano_client.client import StoppableThread
from exoedge.channel import DataOutChannel
from exoedge import logger

LOG = logger.getLogger(__name__)


class ExoEdgeConfig(StoppableThread):
    """
        Base class for all config_* resources.
     """
    def __init__(self, **kwargs):
        StoppableThread.__init__(
            self,
            name=kwargs.get('name', self.__class__.__name__)
        )
        self._lock = threading.Lock()
        self.event_new_config = threading.Event()
        self.config = kwargs.get('config')
        self.config_file = kwargs.get('config_file')
        self.device = kwargs.get('device')
        assert self.device,\
            "{} constructor needs murano_client.client.Client object."\
            .format(self.__class__.__name__)
        self.acktime = 0.0

    def ack_config(self):
        """ ACK config to Murano. """
        with self._lock:
            status, result = self.device.ack(
                resource=self.resource,
                payload=json.dumps(self.config)
            )
            if not status:
                LOG.critical('Ack config failed', result)
            self.acktime = time.time()

    def set_config(self, config):
        """
            Update internal and local config.
            Set's event_new_config.
        """
        if isinstance(config, dict):
            LOG.critical(
                'Received new config:\n{}\n'
                .format(json.dumps(config, indent=2))
            )

            with self._lock:
                self.config = config
            self.write_local_config()
            self.ack_config()
            self.event_new_config.set()
        else:
            LOG.critical(
                'Not setting improper config:\n{}'
                .format(config)
            )

    def read_local_config(self):
        """
            Returns dictionary of config from local file
            if it exists.
        """
        infostr = 'READING LOCAL CONFIG_IO file: {}'.format(
            self.config_file)
        LOG.critical('\n{:-^80}\n'.format(infostr))

        if self.config_file:
            try:
                return json.load(open(self.config_file, 'r'))
            except IOError:
                raise NoLocalConfigFile(
                    'config_io file {} does not exist.'
                    .format(self.config_file))
            except ValueError:
                raise InvalidLocalConfigFile(
                    'config_io file contains invalid JSON: \n{}.'
                    .format(open(self.config_file, 'r').read()))
        else:
            raise NoLocalConfigFile('A local config_io file has not been set.')

    def write_local_config(self):
        """
            Writes the current contents of self.config to
            the configured local filename.
        """
        with self._lock:
            if isinstance(self.config_file, str):
                LOG.info(
                    "Writing {!r}:\n{}"
                    .format(self.config_file, json.dumps(self.config, indent=2))
                )
                with open(self.config_file, 'w') as __f:
                    json.dump(self.config, __f)
            else:
                raise InvalidConfigFileName(
                    "invalid config_file name: {}"
                    .format(self.config_file)
                )

class ConfigWatcher(StoppableThread):
    """
       This diagram illustrates that ExoEdge will continuously monitor
       configuration resources (e.g. config_io, config_applications)
       for updates. The order of precedence is that the remote config
       is first, then local configs. If using ExoEdge when is it to
       be configured with a local config, then it is imperative that
       it connects for the first time where the config resource has
       no value (i.e. it has never been set by ExoSense).

            +-----------------+
            |   while true    |
            +-----------------+
                     |
        +----------------------------<------------+
        |            v                            ^
        |       +----|---+         +--------+     |
        |       | Remote |   N     | Local  | N   |
        ^       | Config |------->-| Config |-->--+
        |       |   ?    |         |   ?    |     ^
        |       +----|---+         +----|---+     |
        |            | Y                | Y       |
        |            |                  v         |
        |            |             +----|---+     |
        |            |             | mtime  |     |
        |            |             |   >    | N   |
        |            |             | acktime|-->--+
        |            |             |   ?    |
        |            |             +----|---+
        |            |                  |
        |            v                  v
        |       +----|---+         +----|---+
        |       | ACK    |         | use    |
        |       | Config |       Y | local  |
        |       |        |<--------| config |
        |       +----|---+         +--------+
        |            |
        |            v
        |     +------|------+
        |     |acktime = now|
        |     +------|------+
        |            |
        |            v
        |        +---|--+
        |        |write |
        ^        |local |
        |        |config|
        |        +---|--+
        |            |
        +--<---------+

    """
    def __init__(self, **kwargs):
        StoppableThread.__init__(self, name='ConfigWatcher')
        self.config_mgrs = {e.name: e for e in kwargs.get('config_mgr_list')}
        self.device = kwargs.get('device')

    def start(self):
        """
            Starts all config manager threads.
        """
        for mgr in self.config_mgrs.values():
            mgr.start()
        super(ConfigWatcher, self).start()

    def stop(self):
        """
            Stops all config manager threads.
        """
        for mgr in self.config_mgrs.values():
            mgr.stop()
        super(ConfigWatcher, self).stop()

    def handle_command(self, inbound, payload):
        LOG.info("ACKing data_out: {}".format(inbound))
        status, result = self.device.ack(
            resource=inbound.resource,
            payload=inbound.payload
        )
        if not status:
            LOG.critical('Ack data_out failed', result)

        LOG.warning("Got data_out: {}".format(inbound))
        for name, value in payload.items():
            LOG.debug("Searching for channel: {}".format(name))
            channel = self.config_mgrs['ConfigIO'].channels.get(name)
            if isinstance(channel, DataOutChannel):
                LOG.info(
                    "Executing data_out {}: {}"
                    .format(name, value)
                )
                channel.execute(value)
            else:
                error_msg = "Channel '{}' not defined as a control channel.".format(name)
                LOG.info(error_msg)
                LOG.debug("\n{}\n".format(channel))

    def check_inbound_queue(self):
        # Check inbound queue for remote config or control command
        inbound = self.device.watch(timeout=1.0)

        if inbound:
            try:
                payload = json.loads(inbound.payload)
            except ValueError:
                LOG.critical(
                    'ERROR: `{}` has invalid JSON payload: {}'
                    .format(inbound, inbound.payload))

            # Match to appropriate manager
            for mgr in self.config_mgrs.values():
                if inbound.resource == mgr.resource:
                    mgr.set_config(payload)
                elif inbound.resource == 'data_out':
                    self.handle_command(inbound, payload)
        else:
            LOG.debug('No remote configs or commands.')

    def run(self):
        """
            Process remote and local configs for all config
            manager threads.

            Process data_out control data.
        """
        # Give a moment for the watch threads to spin up
        time.sleep(1)

        # Init - check inbound once per mgr, ensuring all cloud-staged changes are processed
        for mgr in self.config_mgrs:
            self.check_inbound_queue()

        while not self.is_stopped():
            # 1st check for remote configs or commands
            self.check_inbound_queue()

            # 2nd check local configs
            LOG.debug('Checking local configs.')
            for mgr in self.config_mgrs.values():
                if not isinstance(mgr, ExoEdgeConfig):
                    continue
                LOG.debug(
                    'Checking for local {}: {}'
                    .format(mgr.name, mgr.config_file)
                )
                if not os.path.exists(mgr.config_file):
                    LOG.info(
                        "Config file does not exist: {}"
                        .format(mgr.config_file)
                    )
                    continue
                if os.path.getmtime(mgr.config_file) > mgr.acktime:
                    config = mgr.read_local_config()
                    if config:
                        LOG.critical(
                            'Found new local {}: {}'
                            .format(mgr.name, mgr.config_file)
                        )
                        mgr.set_config(config)
                    else:
                        LOG.debug(
                            "Failed to load new local {}: {}"
                            .format(mgr.name, mgr.config_file)
                        )
                else:
                    LOG.debug("Config '{}' already ACK'ed.".format(mgr.name))

class ExoEdgeException(Exception):
    """ Base exception class for ExoEdge. """
    pass

class NoLocalConfigFile(ExoEdgeException):
    """ ExoEdge Exception: local config file not found. """
    pass

class InvalidLocalConfigFile(ExoEdgeException):
    """ ExoEdge Exception: local config file cannot be parsed. """
    pass

class InvalidConfigFileName(ExoEdgeException):
    """ ExoEdge Exception: local config file has invalid name. """
    pass




"""
Installer for ExoSense Client
"""
import os
from setuptools import setup, find_packages
from exoedge import __version__

DOCS_URL = 'https://github.com/exosite/lib_exoedge_python'

INSTALL_REQUIRES = [
    'appdirs',
    'docopt>=0.6.2',
    'jsonschema==3.2.0',
    'pureyaml',
    'murano-client>=19.5.8',
    'ruamel.yaml',
    'six',
    'urllib3==1.26.7'
]


def read(fname):
    """ Primarily used to open README file. """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


try:
    README = read('README.rst')
except:
    README = ''

setup(
    name="exoedge",
    version=__version__,
    author="Exosite LLC",
    author_email="support@exosite.com",
    description="""ExoEdge is the Python library for interacting with Exosite's ExoSense Industrial IoT Solution.""",
    license="Apache 2.0",
    keywords="murano exosite iot iiot gateway edge exoedge exosense",
    url="https://github.com/exosite/lib_exoedge_python",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'edged = exoedge.edged:main'
        ]
    },
    install_requires=INSTALL_REQUIRES,
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Internet",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
    ],
    data_files=[],
    package_data={'exoedge': ['config_io_schema.yaml']}
    )



# pylint: disable=C0103,W0602,C0325,C0301,C0111
import os
import unittest
import json
from exoedge.config_io import ConfigIO
from exoedge.config_applications import ConfigApplications
from murano_client.client import MuranoClient

test_dir = os.path.dirname(os.path.abspath(__file__))


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass
    @classmethod
    def tearDownClass(cls):
        pass
    def setUp(self):
        self.test_name = self.id().split('.')[-1]
        self.test_config_file = os.path.join(test_dir, self.test_name)
        with open(self.test_config_file, 'w'):
            pass
        print("setup for test: {}".format(self.test_name))

    def tearDown(self):
        print("teardown for test: {}".format(self.test_name))
        os.unlink(self.test_config_file)

    def test_001_config_io_default(self):

        config_io = ConfigIO(
            device=MuranoClient(
                murano_host="https://dne.m2.exosite.io/",
                watchlist=['data_out']
            )
        )
        self.assertEqual(config_io.config_file, 'config_io.json')

    def test_002_config_io_configured(self):

        config_io = ConfigIO(
            device=MuranoClient(
                murano_host="https://dne.m2.exosite.io/",
                watchlist=['data_out']
            ),
            config_io_file='/custom/path.json'
        )
        self.assertEqual(config_io.config_file, '/custom/path.json')

    def test_003_set_config_verify_file_sync(self):

        config_io = ConfigIO(
            device=MuranoClient(
                murano_host="https://dne.m2.exosite.io/",
                watchlist=['data_out']
            ),
            config_io_file=self.test_config_file
        )
        self.assertEqual(config_io.config_file, self.test_config_file)

        config = json.load(open(os.path.join(test_dir, 'assets/complete-channel-config.json')))
        config_io.set_config(config)
        self.assertEqual(config_io.config, config)
        self.assertEqual(config, json.load(open(self.test_config_file, 'r')))







    def test_010_config_applications_default(self):

        config_applications = ConfigApplications(
            device=MuranoClient(
                murano_host="https://dne.m2.exosite.io/",
                watchlist=['data_out']
            )
        )
        self.assertEqual(config_applications.config_file, 'config_applications.json')

    def test_020_config_applications_configured(self):

        config_applications = ConfigApplications(
            device=MuranoClient(
                murano_host="https://dne.m2.exosite.io/",
                watchlist=['data_out']
            ),
            config_applications_file='/custom/path.json'
        )
        self.assertEqual(config_applications.config_file, '/custom/path.json')

    def test_030_set_config_verify_file_sync(self):

        config_applications = ConfigApplications(
            device=MuranoClient(
                murano_host="https://dne.m2.exosite.io/",
                watchlist=['data_out']
            ),
            config_applications_file=self.test_config_file
        )
        self.assertEqual(config_applications.config_file, self.test_config_file)

        config = json.load(open(os.path.join(test_dir, 'assets/complete-channel-config.json')))
        config_applications.set_config(config)
        self.assertEqual(config_applications.config, config)
        self.assertEqual(config, json.load(open(self.test_config_file, 'r')))








def main():
    unittest.main()

if __name__ == "__main__":
    main()



import json
import os
import re
import sqlite3
import threading
import time

import six.moves.urllib as urllib
from appdirs import user_data_dir

from exoedge import logger

LOG = logger.getLogger(__name__)

# Murano has payload limit to 64KB
PAYLOAD_MAX_SIZE = 64000

def murano_urlencode(data):
    return urllib.parse.quote_plus(data.encode('utf-8'))


class Queuing(object):
    def __init__(self, host, device_id, device,
                 batch_amount=0,
                 db_max_size=20000000,  # 20 MB
                 db_path=user_data_dir('exoedge', 'exosite'),
                 drop_strategy='oldest',
                 record_rate=1,
                 remove_unused_device=False
                 ):
        self.batch_amount = batch_amount
        self.record_rate = record_rate
        self.dequeue_enabled = False
        self.table_name = 'queue.' + self.product_id(host) + '.' + device_id
        self.db_path = os.path.normpath(db_path)
        self.device = device
        self.device.set_http_timeout(10)
        self.db = DB(self.db_path, self.table_name, db_max_size, drop_strategy)

        if not self.db.is_table_exist():
            self.db.create_table()

        self.is_table_empty = True if self.db.is_table_empty() else False
        if not self.is_table_empty:
            self.enable_dequeue()

        if remove_unused_device:
            self.db.delete_unused_table()

    def enable_dequeue(self):
        self.dequeue_enabled = True
        self.thread = threading.Thread(target=self.dequeue_table)
        self.thread.start()

    def disable_dequeue(self):
        LOG.info('Stoping data queuing...')
        self.dequeue_enabled = False
        if hasattr(self, 'thread'):
            self.thread.join()

    def dequeue_table(self):
        while self.dequeue_enabled:
            records = self.db.read_table(
                auto=(self.batch_amount == 0), amount=self.batch_amount)

            if not len(records):
                self.is_table_empty = True
                self.dequeue_enabled = False
                continue

            try:
                if self.device.outbound_protocol == 'https':
                    self.https(records)
                else:
                    self.mqtt(records)

                self.db.table.delete(start=records[0][0], end=records[-1][0])
            except Exception as err:
                LOG.warn('Consume record failed')

            time.sleep(self.record_rate)

    def https(self, records):
        data = {}
        for row in records:
            data[row[1]] = row[2]
        payloads = {"data_in": data}

        status, result = self.device.tell(payloads=payloads)

        if not status:
            raise Exception('Https recored fail')

    def mqtt(self, records):
        payload = json.dumps(list(map(lambda row: {
            'timestamp': row[1],
            'values': {
                'data_in': row[2]
            }
        }, records)))
        status, result = self.device.tell(payloads=payloads)

        if not status:
            raise Exception('Mqtt not published')

    def product_id(self, host):
        """
            ex:
            host = "mqtt://z4fwqr3fmr0200000.m2.exosite.io/"
            return z4fwqr3fmr0200000
        """

        try:
            return re.search(r"(?<=//)\w+", host).group(0)
        except AttributeError as err:
            LOG.error('Invalid host fomat: {}'.format(err))
            raise

    def tell(self, resource=None, timestamp=None, payload=None):
        """
            resource: device's resource.
            timestamp: time.time()
            payload: json string
        """

        def payload_size_over_limit():
            payload_size = self.db.data_size(payload)
            if payload_size>=PAYLOAD_MAX_SIZE:
                LOG.warn('Tell failed, payload size {} byte and limist is {} byte'.format(payload_size, PAYLOAD_MAX_SIZE))
                return True
            else:
                return False

        if self.is_table_empty:
            status, result = self.device.tell(
                resource=resource,
                timestamp=timestamp,
                payload=payload
            )

            if not status:
                if payload_size_over_limit():
                    return False, 'payload size over limit {}'.format(PAYLOAD_MAX_SIZE)
                res = self.db.insert_record(timestamp, payload)
                if res is not None:
                    return False, 'Insert db fail: {}'.format(res)
                self.is_table_empty = False
                self.enable_dequeue()
                return True, 'Tell failed, putting payload into DB'
            return status, result
        else:
            if payload_size_over_limit():
                return False, 'payload size over limit {}'.format(PAYLOAD_MAX_SIZE)
            res = self.db.insert_record(timestamp, payload)
            if res is not None:
                return False, 'Insert db fail: {}'.format(res)
            return True, 'DB is not empty, putting payload into DB'


class Table(object):
    def __init__(self, query_db, name):
        self.name = name
        self.query_db = query_db

    def read(self, limit=1, startId=0):
        res = self.query_db(
            """
                SELECT * FROM '{}' WHERE id >= {} LIMIT {}
            """.format(self.name, startId, limit),
            one=False
        )
        return res

    def insert(self, time, value):
        res = self.query_db(
            """
                INSERT INTO '{}' VALUES (null, '{}', '{}')
            """.format(self.name, time, value)
        )
        return res

    def delete(self, start=0, end=99999999):
        res = self.query_db(
            """
               DELETE FROM '{}' WHERE id between {} and {}
            """.format(self.name, start, end)
        )
        return res

    def drop_oldest_records(self, drop_range_seconds=600):
        """
           Check all of table to delete the oldest 10 mins data, except the last 10 mins data.
        """
        table = self.the_oldest_record_table()
        if not table:
            return False

        oldest_time = table[1]

        if oldest_time + drop_range_seconds > time.time():
            return False

        res = self.query_db(
            """
                DELETE FROM '{}'
                WHERE timestamp <= '{}'
            """.format(table[0], oldest_time + drop_range_seconds)
        )
        return res

    def the_oldest_record_table(self):
        """
            return ex: ('queue.abcdcdkk2ls00000.device', 1.0)
        """
        oldest_table_time = 9999999999
        table = ()

        for table_name in self.all_table_name():
            timestamp = self.query_db(
                """
                    SELECT timestamp FROM '{}' ORDER BY timestamp ASC limit 1
                """.format(table_name)
            )
            if timestamp and timestamp < oldest_table_time:
                oldest_table_time = timestamp
                table = (table_name, oldest_table_time)

        return table

    def all_table_name(self):
        """
           return ex: ['queue.abcdcdkk2ls00000.device', 'queue.abcdcdkk2ls00000.device2']
        """
        res = self.query_db(
            """
                SELECT name FROM sqlite_master where name LIKE 'queue.%'
            """, one=False
        )
        return [i[0] for i in res]

    def delete_table(self, table_name):
        res = self.query_db(
            """
                DROP TABLE IF EXISTS '{}'
            """.format(table_name), one=False
        )

        return res


class DB(Table):
    def __init__(self, db_path, table_name, db_max_size, drop_strategy):
        self.table = Table(self.query_db, table_name)

        if not os.path.isdir(db_path):
            os.makedirs(db_path)
        self.path = os.path.join(db_path, 'exoedge.db')
        self.table_name = table_name
        self.db_max_size = db_max_size
        self.drop_strategy = drop_strategy
        self.is_vacuumed = True

    def insert_record(self, time, value):
        real_size = os.stat(self.path).st_size

        if real_size >= self.db_max_size:
            LOG.warn('DB current size is {} byte, max size is {} byte'.format(
                real_size, self.db_max_size))
            self.is_vacuumed = False

            if self.drop_strategy == 'oldest':
                # The drop size may small then insert size so may over size limitation little bit.
                self.drop_oldest_records()
                return None
            elif self.drop_strategy == 'latest':
                return None
            else:
                return 'Drop strategy not supported'

        return self.table.insert(time, value)

    def delete_unused_table(self):
        for table_name in self.table.all_table_name():
            if table_name != self.table_name:
                self.table.delete_table(table_name)

        return True

    def read_table(self, auto=True, amount=0):
        """
            vacuum DB when DB from full to empty.
        """
        res = self.limit_data_size(auto, amount)

        if not len(res) and not self.is_vacuumed:
            LOG.info('Doing DB vacuum')
            self.vacuum()
            self.is_vacuumed = True
        return res

    def query_db(self, query, values=None, one=True):
        """
            Method for all queries to the database.

            query: an SQL statement string
            values: a tuple of query values for
                    statements like SELECT, UPDATE, etc.
                    number of values in tuple should match
                    number of '?' in query
            one: default behavior for query return value
                 is to call and return query.fetchone().
                 if one=False, return query.fetchall().
        """

        LOG.debug("Executing query: {!r}, {}, one = {}"
                  .format(query, values, one))
        result = None

        with sqlite3.connect(self.path) as conn:
            # if db not exist, will auto created
            cursor = conn.cursor()
            if values:
                cursor.execute(query, values)
            else:
                cursor.execute(query)
            conn.commit()
            result = cursor.fetchone() if one else cursor.fetchall()
            if result and one:
                result = result[0]
        LOG.debug("db result = {0}".format(result))
        return result

    def create_table(self):
        self.query_db(
            """
                CREATE TABLE '{}' (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    timestamp REAL,
                    data TEXT
                )
            """.format(self.table_name)
        )
        LOG.info('Table {} created'.format(self.table_name))

    def is_table_empty(self):
        res = self.query_db(
            """
                SELECT * FROM '{}'
            """.format(self.table_name)
        )
        return False if res else True

    def is_table_exist(self):
        """
         res example: u'z4fwqr3fmr0200000.j1939'
        """

        res = self.query_db(
            """
                SELECT name FROM sqlite_master WHERE type='table' AND name='{}';
            """.format(self.table_name)
        )
        return True if res else False

    def vacuum(self):
        res = self.query_db("VACUUM")
        return res

    def limit_data_size(self, auto=True, amount=0):
        """
            When send data Murano Resources, the maximum size is 64K
            auto: True - auto retrieve data which close to 64K, ignore amount parameter
                  False - retrieve data by given amount and deduct to 64K
        """
        # byte
        size_limit = PAYLOAD_MAX_SIZE
        size = 0

        if auto:
            records = []
            start_id = 0

            while True:
                data = self.table.read(startId=start_id, limit=1)
                if not data:
                    break
                size += self.data_size('{timestamp:.7f}={generateddata}&'.format(timestamp=data[0][1], generateddata=data[0][2]))
                if size > size_limit:
                    break
                records.append(data[0])
                start_id = data[0][0]+1
        else:
            records = self.table.read(amount)

            for index, record in enumerate(records):
                size += self.data_size('{timestamp:.7f}={generateddata}&'.format(timestamp=record[1], generateddata=record[2]))
                if size > size_limit:
                    return records[:index]

        return records

    def data_size(self, data):
        return len(murano_urlencode(data))



# pylint: disable=C1801,W1202,C0103

import sys
import traceback
import logging
import logging.config
from logging.handlers import RotatingFileHandler
from exoedge.constants import DEFAULTS
from murano_client.logger import LOG_FORMAT, FORMATTER

if not hasattr(sys.modules.get(__name__), 'LOGGERS'):
    global LOGGERS
    LOGGERS = []

LOG_KWARGS = {
    'format': LOG_FORMAT,
    'level': getattr(logging, DEFAULTS['debug'].upper())
}

if DEFAULTS['log-filename'] in ['stdout', 'stderr']:
    DEFAULTS['log-filename'] = getattr(sys, DEFAULTS['log-filename'])
    LOG_KWARGS['stream'] = DEFAULTS['log-filename']
else:
    LOG_KWARGS['filename'] = DEFAULTS['log-filename']
    LOG_KWARGS['filemode'] = 'a'

logging.basicConfig(**LOG_KWARGS)
logging.getLogger().propagate = False
logging.getLogger().setLevel(getattr(logging, DEFAULTS['debug'].upper()))
LOGGERS.append(logging.getLogger())

def log_exceptions(ex_cls, ex, tb):
    _tb = ''
    for e in traceback.format_tb(tb):
        _tb += e
    logging.critical('edged logging unhandled exception: \n{}{}'
                     .format(_tb, ex))
sys.excepthook = log_exceptions

def getLogger(*args, **kwargs):
    global LOGGERS
    name = args[0]
    if not name.startswith("exoedge"):
        name = 'exoedge.' + name
    log = logging.getLogger(name=name)
    if kwargs.get('level') != None and isinstance(kwargs.get('level'), int):
        level = kwargs.get('level') or logging.DEBUG
    else:
        level = LOG_KWARGS.get('level')
    log.setLevel(level)
    log.info("Starting logger {}({}): {}".format(name, log.level, log))
    if not len(log.handlers) and DEFAULTS['log-filename'] not in [sys.stdout, sys.stderr]:
        rfh = RotatingFileHandler(
            kwargs.get('log-filename') or DEFAULTS['log-filename'],
            maxBytes=kwargs.get('log-max-bytes') or 1024*1000,
            backupCount=kwargs.get('log-max-backups') or 3,
        )
        rfh.setFormatter(FORMATTER)
        log.addHandler(rfh)

    elif not len(log.handlers) and DEFAULTS['log-filename'] in [sys.stdout, sys.stderr]:
        streamh = logging.StreamHandler(DEFAULTS['log-filename'])
        streamh.setFormatter(FORMATTER)
        log.addHandler(streamh)

    log.propagate = False
    LOGGERS.append(log)
    return log

def getExoEdgeLoggers():
    global LOGGERS
    return LOGGERS


#pylint: disable=W1202
"""
Usage:
    edged [options] [<command>] [<args>...]

Commands:
    go          Start recording to ExoSense

Options:
    -h --help                               Show this screen.
    -v --version                            Print the current version of the ExoEdge library.
    -L --list-commands                      Print installed and available commands.
    -i --ini-file <file>                    INI file with device information.
    -c --config-io-file <cfgfile>           Local file in which to cache config_io, or to start
                                            ExoEdge with a config_io setting for the first time.
                                            Any changes to this setting in ExoSense will be
                                            synced and cached in this file.
    -a --config-applications-file <cfgfile> Local file in which to cache config_applications, or
                                            to start ExeEdge with a config_applications setting
                                            for the first time. Any changes to this setting in
                                            ExoSense will be synced and cached in this file.
    -s --murano-id <sn>                     The device serial number to use.
    -t --murano-token <token>               Token for device authentication.
    -K --pkeyfile <pkey>                    Private key for TLS provisioning.
    -C --certfile <cert>                    Public cert for TLS provisioning.
    -E --murano-cacert <cacert>             CA cert for PKI integration.
    -H --murano-host <host>                 Set host for API requests.
    -p --murano-port <port>                 Set port for API requests.
    -d --debug <lvl>                        Tune the debug output. Logs curl commands at
                                            DEBUG.
                                            (DEBUG|INFO|WARNING|ERROR|CRITICAL).
    -q --queuing <bool>                     Store data when no network and resend data when backing online.
                                            Default: True (True|False)
    --queuing-batch-amount <number>         For queuing. Numbers of records to send in one payload in one payload.
                                            When the number is 0, it would auto-adjust payload to within and close to 64KB.
                                            Default: 0
    --queuing-db-max-size <byte>            For queuing. Limit DB file size to prevent full fill hard disk.
                                            Default: 20000000; 20MB
    --queuing-db-path <path>                For queuing. Put exoedge.db at a specific path.
                                            Default:
                                                On macOS: '/Users/trentm/Library/Application Support/exoedeg'
                                                On Windows: 'C:\\Users\\trentm\\AppData\\Local\\Acme\\exoedeg'
                                                On Linux: '/home/trentm/.local/share/exoedeg
    --queuing-drop-strategy <oldest|latest> For queuing. The behavior when DB is full. Drop old data and then append new data or just drop new data.
                                            Default: oldest
    --queuing-record-rate <second>          For queuing. How long report data one time.
                                            Default: 1
    --queuing-remove-unused-device <bool>   Remove DB's table when device is not used. The device is according to --murano-id. This is useful when you change devices and want to clean DB.
                                            Default: False
    --strategy <strategy>                   DEPRECATED. See Notes #2 below for details.
    --local-strategy                        DEPRECATED. See option --strategy.
    --no-filesystem                         DEPRECATED. Don't rely on a file system.
    --no-config-cache                       DEPRECATED. Don't store a local copy of config_io.
    --no-config-sync                        DEPRECATED. Don't keep config_io synced with ExoSense.
    --http-timeout <timeout>                Timeout to use for long-poll requests.
    --edged-timeout <timeout>               Timeout for edged process, in seconds. The edged
                                            process will exit after <timeout> seconds.
    --watchlist <watchlist>                 DEPRECATED. Murano resources to watch. Comma separated
                                            list (e.g. --watchlist=config_io,data_out). Settable in
                                            INI file.
    <command>                               The ExoEdge subcommand name.
    <args>                                  Supported arguments for <command>

Notes:

    1. ExoEdge logs default to 'stdout'. This can be overridden to either 'stderr' or
       a logfile of choice. If using a logfile of choice, then the log file will be
       rotated once it reaches a given size (in bytes).

       e.g.

            $ export EDGED_LOG_FILENAME=${PWD}/edged.log
            $ edged -i f5330e5s8cho0000.ini go

       Other supported environment variables for logging configuration are:

            EDGED_LOG_DEBUG (default:CRITICAL)
            EDGED_LOG_MAX_BYTES (default:1024000)
            EDGED_LOG_MAX_BACKUPS (default:3)

    2. The ExoEdge config management state machine* replaces the
       deprecated --strategy CLI options.

       * See exoedge.configs.ConfigWatcher docstring for diagram.

"""
from __future__ import print_function
import sys
import logging
import os
import pkgutil
from docopt import docopt, DocoptExit
import exoedge.commands
from exoedge import __version__ as VERSION
from exoedge import logger

LOG = logger.getLogger(__name__)

def main():
    args = docopt(__doc__, version=VERSION, options_first=True)
    # If no command specified, fallback to 'go' (quickly enables support for pre-built executables)
    args['<command>'] = args['<command>'] or 'go'

    if args.get('--version'):
        print(VERSION)
        return

    if args.get('--list-commands'):
        sys.path.insert(0, os.path.join(os.path.abspath(os.path.dirname(__file__)), 'commands'))
        for _, module_name, _ in pkgutil.iter_modules(exoedge.commands.__path__):
            print(module_name)
        return

    if args.get('--debug'):
        level = args.get('--debug')
        if isinstance(level, str):
            level = level.upper()
        LOG.setLevel(getattr(logging, level) or logging.CRITICAL)

    arg = '--local-strategy'
    if args.get(arg):
        LOG.warning(
            "DEPRECATED Command Line Option: <{}>. Please use --strategy option instead (see --help for more info)."
            .format(arg))

    arg = '--strategy'
    if args.get(arg):
        LOG.warning(
            "DEPRECATED Command Line Option: <{}>. See edged -h for more info on configuration management."
            .format(arg))

    for arg in ['--no-filesystem', '--no-config-cache', '--no-config-sync', '--watchlist']:
        if args.get(arg):
            LOG.warning(
                "DEPRECATED Command Line Option: <{}>. Option removed."
                .format(arg))

    if args.get('<command>'):
        command_name = args.pop('<command>')
        argv = args.pop('<args>')
        if argv is None:
            argv = {}

        sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'commands'))
        try:
            # TODO: If more commands ever added, would need refactoring for dynamic loading
            if command_name != 'go':
                raise ImportError

            from exoedge.commands import go
            command_class = getattr(go, 'ExoCommand')
        except ImportError as exc:
            raise DocoptExit("Cannot find command {!r}: {}"
                             .format(command_name, exc))

        command = command_class(argv, args)

        command.execute()
        return

    raise DocoptExit("Provide an option or subcommand\n")


if __name__ == '__main__':
    main()
